"""Tests for MongoDBUpsertManyNode."""

from __future__ import annotations
from typing import Any
from unittest.mock import Mock
import pytest
from langchain_core.runnables import RunnableConfig
from pymongo.results import BulkWriteResult
from orcheo.graph.state import State
from orcheo.nodes.mongodb import MongoDBUpsertManyNode
from tests.nodes.conftest import MongoTestContext


_RECORDS = [
    {
        "link": "https://example.com/1",
        "title": "One",
        "read": True,
        "fetched_at": "2026-02-28T09:00:00+00:00",
    },
    {
        "link": "https://example.com/2",
        "title": "Two",
        "read": True,
        "fetched_at": "2026-02-28T09:30:00+00:00",
    },
]


def _base_state() -> State:
    return State(messages=[], inputs={}, results={})


def _state_with_records(
    *,
    source_key: str = "fetch_rss",
    records_field: str = "documents",
    records: list[dict[str, Any]] | None = None,
    source_as_list: bool = False,
) -> State:
    payload = list(records) if records is not None else list(_RECORDS)
    if source_as_list:
        results: dict[str, Any] = {source_key: payload}
    else:
        results = {source_key: {records_field: payload}}
    return State(messages=[], inputs={}, results=results)


def _mock_bulk_write_result() -> Mock:
    result = Mock(spec=BulkWriteResult)
    result.inserted_count = 0
    result.matched_count = 1
    result.modified_count = 1
    result.deleted_count = 0
    result.upserted_count = 1
    result.upserted_ids = {1: "507f1f77bcf86cd799439011"}
    result.acknowledged = True
    return result


def _build_node(**overrides: Any) -> MongoDBUpsertManyNode:
    defaults: dict[str, Any] = {
        "name": "upsert_many",
        "database": "test_db",
        "collection": "rss_feeds",
        "source_result_key": "fetch_rss",
        "filter_fields": ["link"],
        "exclude_fields": ["read"],
        "set_on_insert": {"read": False},
    }
    defaults.update(overrides)
    return MongoDBUpsertManyNode(**defaults)


def test_upsert_many_resolve_records_returns_empty_for_non_mapping_results() -> None:
    node = _build_node()
    state = State(messages=[], inputs={}, results="invalid")

    assert node._resolve_records(state) == []


def test_upsert_many_resolve_records_returns_empty_for_non_list_documents() -> None:
    node = _build_node()
    state = State(
        messages=[],
        inputs={},
        results={"fetch_rss": {"documents": "invalid"}},
    )

    assert node._resolve_records(state) == []


def test_upsert_many_resolve_records_returns_empty_for_non_mapping_record() -> None:
    node = _build_node()
    state = State(
        messages=[],
        inputs={},
        results={
            "fetch_rss": {"documents": [{"link": "https://example.com/1"}, "bad"]}
        },
    )

    assert node._resolve_records(state) == []


@pytest.mark.asyncio
async def test_upsert_many_builds_bulk_operations(
    mongo_context: MongoTestContext,
) -> None:
    mongo_context.collection.bulk_write.return_value = _mock_bulk_write_result()
    node = _build_node()

    result = await node.run(_state_with_records(), RunnableConfig())

    mongo_context.collection.bulk_write.assert_called_once()
    operations = mongo_context.collection.bulk_write.call_args[0][0]
    assert len(operations) == 2
    assert operations[0]._filter == {"link": "https://example.com/1"}
    assert operations[0]._doc == {
        "$set": {
            "link": "https://example.com/1",
            "title": "One",
            "fetched_at": "2026-02-28T09:00:00+00:00",
        },
        "$setOnInsert": {"read": False},
    }
    assert operations[0]._upsert is True

    assert result["data"]["operation"] == "bulk_write"
    assert result["data"]["matched_count"] == 1
    assert result["data"]["modified_count"] == 1
    assert result["data"]["upserted_count"] == 1
    assert result["data"]["upserted_ids"] == {"1": "507f1f77bcf86cd799439011"}


@pytest.mark.asyncio
async def test_upsert_many_honors_set_fields(
    mongo_context: MongoTestContext,
) -> None:
    mongo_context.collection.bulk_write.return_value = _mock_bulk_write_result()
    node = _build_node(set_fields=["title"])

    await node.run(_state_with_records(), RunnableConfig())

    operations = mongo_context.collection.bulk_write.call_args[0][0]
    assert operations[0]._doc == {
        "$set": {"title": "One"},
        "$setOnInsert": {"read": False},
    }


@pytest.mark.asyncio
async def test_upsert_many_accepts_list_source(
    mongo_context: MongoTestContext,
) -> None:
    mongo_context.collection.bulk_write.return_value = _mock_bulk_write_result()
    node = _build_node()

    await node.run(_state_with_records(source_as_list=True), RunnableConfig())

    operations = mongo_context.collection.bulk_write.call_args[0][0]
    assert len(operations) == 2


@pytest.mark.asyncio
async def test_upsert_many_requires_records(
    mongo_context: MongoTestContext,
) -> None:
    node = _build_node(source_result_key="missing")

    with pytest.raises(ValueError, match="No records available to upsert"):
        await node.run(_base_state(), RunnableConfig())


@pytest.mark.asyncio
async def test_upsert_many_requires_filter_fields(
    mongo_context: MongoTestContext,
) -> None:
    node = _build_node(filter_fields=[])

    with pytest.raises(
        ValueError,
        match="filter_fields must contain at least one field",
    ):
        await node.run(_state_with_records(), RunnableConfig())


@pytest.mark.asyncio
async def test_upsert_many_rejects_missing_filter_field(
    mongo_context: MongoTestContext,
) -> None:
    node = _build_node()
    state = _state_with_records(records=[{"title": "Missing link"}])

    with pytest.raises(ValueError, match="missing filter field 'link'"):
        await node.run(state, RunnableConfig())


@pytest.mark.asyncio
async def test_upsert_many_rejects_empty_generated_update(
    mongo_context: MongoTestContext,
) -> None:
    node = _build_node(set_fields=["missing_field"], set_on_insert={})

    with pytest.raises(ValueError, match="Generated update document must not be empty"):
        await node.run(_state_with_records(), RunnableConfig())



"""Utilities for normalising graph configuration structures."""

from __future__ import annotations
from collections.abc import Iterable, Mapping
from typing import Any
from langgraph.graph import END, START


def normalise_edges(edges: Iterable[Any]) -> list[tuple[str, str]]:
    """Return list of (source, target) tuples for the given edge entries."""
    normalised: list[tuple[str, str]] = []
    for entry in edges:
        if isinstance(entry, Mapping):
            source = entry.get("source")
            target = entry.get("target")
        else:
            try:
                source, target = entry  # type: ignore[misc]
            except (TypeError, ValueError) as exc:  # pragma: no cover - defensive
                msg = f"Invalid edge entry: {entry!r}"
                raise ValueError(msg) from exc

        if not isinstance(source, str) or not isinstance(target, str):
            msg = f"Edge endpoints must be strings: {entry!r}"
            raise ValueError(msg)
        normalised.append((source, target))
    return normalised


def normalise_vertex(name: str) -> Any:
    """Map sentinel vertex names to LangGraph constants."""
    if name == "START":
        return START
    if name == "END":
        return END
    return name



"""Workflow listener service helpers."""

from __future__ import annotations
from collections.abc import Mapping
from typing import Any
from orcheo_sdk.cli.http import ApiClient


def pause_workflow_listener_data(
    client: ApiClient,
    workflow_id: str,
    subscription_id: str,
    *,
    actor: str,
) -> dict[str, Any]:
    """Pause one workflow listener subscription."""
    return _update_workflow_listener_status_data(
        client,
        workflow_id,
        subscription_id,
        action="pause",
        actor=actor,
    )


def resume_workflow_listener_data(
    client: ApiClient,
    workflow_id: str,
    subscription_id: str,
    *,
    actor: str,
) -> dict[str, Any]:
    """Resume one workflow listener subscription."""
    return _update_workflow_listener_status_data(
        client,
        workflow_id,
        subscription_id,
        action="resume",
        actor=actor,
    )


def _update_workflow_listener_status_data(
    client: ApiClient,
    workflow_id: str,
    subscription_id: str,
    *,
    action: str,
    actor: str,
) -> dict[str, Any]:
    payload = client.post(
        f"/api/workflows/{workflow_id}/listeners/{subscription_id}/{action}",
        json_body={"actor": actor},
    )
    if isinstance(payload, Mapping):
        return dict(payload)
    return {}


__all__ = [
    "pause_workflow_listener_data",
    "resume_workflow_listener_data",
]



"""Workflow upload validation and formatting tests."""

from __future__ import annotations
import sys
from pathlib import Path
import httpx
import pytest
import respx
from typer.testing import CliRunner
from orcheo_sdk.cli.errors import CLIError
from orcheo_sdk.cli.main import app
from orcheo_sdk.cli.workflow import _load_workflow_from_python


def test_load_workflow_from_python_missing_workflow_variable(
    runner: CliRunner, env: dict[str, str], tmp_path: Path
) -> None:
    """Test loading Python file without 'workflow' treats it as LangGraph script."""
    py_file = tmp_path / "no_workflow.py"
    py_file.write_text("some_other_var = 123", encoding="utf-8")

    # Now it treats files without 'workflow' as LangGraph scripts
    # and tries to create a workflow and ingest them
    created_workflow = {"id": "wf-new", "name": "no-workflow"}
    # The ingestion will fail because it's not valid LangGraph code
    with respx.mock() as router:
        router.post("http://api.test/api/workflows").mock(
            return_value=httpx.Response(201, json=created_workflow)
        )
        router.post("http://api.test/api/workflows/wf-new/versions/ingest").mock(
            return_value=httpx.Response(
                400, json={"detail": "Script did not produce a LangGraph StateGraph"}
            )
        )
        result = runner.invoke(
            app,
            ["workflow", "upload", str(py_file)],
            env=env,
        )
    assert result.exit_code != 0
    assert isinstance(result.exception, CLIError)
    assert "Failed to ingest LangGraph script" in str(result.exception)


def test_load_workflow_from_python_wrong_type(
    runner: CliRunner, env: dict[str, str], tmp_path: Path
) -> None:
    """Test loading Python file with wrong workflow type fails."""
    py_file = tmp_path / "wrong_type.py"
    py_file.write_text("workflow = 'not a Workflow instance'", encoding="utf-8")

    result = runner.invoke(
        app,
        ["workflow", "upload", str(py_file)],
        env=env,
    )
    assert result.exit_code != 0
    assert isinstance(result.exception, CLIError)
    assert "must be an orcheo_sdk.Workflow instance" in str(result.exception)


def test_load_workflow_from_python_adds_managed_plugin_site_packages(
    monkeypatch: pytest.MonkeyPatch, tmp_path: Path
) -> None:
    """Managed plugin installs should be importable during local workflow loading."""
    plugin_dir = tmp_path / "plugins"
    site_packages = (
        plugin_dir
        / "venv"
        / "lib"
        / f"python{sys.version_info.major}.{sys.version_info.minor}"
        / "site-packages"
    )
    module_dir = site_packages / "orcheo_plugin_fixture_runtime"
    module_dir.mkdir(parents=True)
    (module_dir / "__init__.py").write_text("VALUE = 'ready'\n", encoding="utf-8")

    workflow_file = tmp_path / "workflow.py"
    workflow_file.write_text(
        "\n".join(
            [
                "from orcheo_plugin_fixture_runtime import VALUE",
                "",
                "class FakeWorkflow:",
                "    def to_deployment_payload(self):",
                "        return {'_type': 'langgraph_script', 'script': VALUE}",
                "",
                "workflow = FakeWorkflow()",
            ]
        ),
        encoding="utf-8",
    )

    monkeypatch.setenv("ORCHEO_PLUGIN_DIR", str(plugin_dir))

    before = list(sys.path)
    try:
        payload = _load_workflow_from_python(workflow_file)
    finally:
        sys.path[:] = before
        sys.modules.pop("orcheo_plugin_fixture_runtime", None)

    assert payload == {"_type": "langgraph_script", "script": "ready"}


def test_load_workflow_from_json_is_rejected(
    runner: CliRunner, env: dict[str, str], tmp_path: Path
) -> None:
    """JSON upload files are rejected before parsing."""
    json_file = tmp_path / "array.json"
    json_file.write_text('["not", "an", "object"]', encoding="utf-8")

    result = runner.invoke(
        app,
        ["workflow", "upload", str(json_file)],
        env=env,
    )
    assert result.exit_code != 0
    assert isinstance(result.exception, CLIError)
    assert "Only .py files are supported" in str(result.exception)


def test_format_workflow_as_python_from_langgraph_script() -> None:
    """Script versions are exported as their original source."""
    from orcheo_sdk.cli.workflow import _format_workflow_as_python

    workflow = {"name": "TestWorkflow"}
    graph = {"format": "langgraph-script", "source": "print('hello')\n"}

    result = _format_workflow_as_python(workflow, graph)
    assert result == "print('hello')\n"


def test_format_workflow_as_python_rejects_non_script_payload() -> None:
    """Non-script versions raise a clear export error."""
    from orcheo_sdk.cli.workflow import _format_workflow_as_python

    with pytest.raises(CLIError, match="unsupported format"):
        _format_workflow_as_python({"name": "legacy"}, {"nodes": [], "edges": []})


def test_format_workflow_as_python_rejects_blank_script_source() -> None:
    """Blank script source should be rejected even for langgraph-script format."""
    from orcheo_sdk.cli.workflow import _format_workflow_as_python

    with pytest.raises(CLIError, match="unsupported format"):
        _format_workflow_as_python(
            {"name": "empty-source"},
            {"format": "langgraph-script", "source": "   "},
        )



from __future__ import annotations
import pytest
from orcheo.plugins import PluginLoadReport, PluginLoadResult
from orcheo_backend.app import plugin_inventory


def test_list_runtime_plugins_merges_inventory_and_load_status(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Runtime inventory reports enabled/loaded details per plugin."""
    fake_rows = [
        {
            "name": "orcheo-plugin-foo",
            "enabled": True,
            "status": "enabled",
            "version": "1.0.0",
            "exports": ["nodes"],
            "source": "orcheo-plugin-foo",
        },
        {
            "name": "orcheo-plugin-bar",
            "enabled": False,
            "status": "installed",
            "version": "2.0.0",
            "exports": [],
            "source": "orcheo-plugin-bar",
        },
    ]

    class StubManager:
        @staticmethod
        def list_plugins() -> list[dict[str, object]]:
            return fake_rows

    monkeypatch.setattr(plugin_inventory, "PluginManager", lambda: StubManager())
    monkeypatch.setattr(
        plugin_inventory,
        "load_enabled_plugins",
        lambda *, force=False: PluginLoadReport(
            generation=1,
            results=[
                PluginLoadResult(name="orcheo-plugin-foo", loaded=True),
            ],
        ),
    )

    inventory = plugin_inventory.list_runtime_plugins()
    assert len(inventory) == 2
    foo = inventory[0]
    assert foo["name"] == "orcheo-plugin-foo"
    assert foo["loaded"] is True
    assert foo["load_error"] is None
    bar = inventory[1]
    assert bar["name"] == "orcheo-plugin-bar"
    assert bar["loaded"] is False
    assert bar["load_error"] is None


def test_missing_required_plugins_filters_disabled_and_unloaded(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Required plugin detection ignores blanks and flags unavailable items."""
    monkeypatch.setattr(
        plugin_inventory,
        "list_runtime_plugins",
        lambda: [
            {
                "name": "orcheo-plugin-foo",
                "enabled": True,
                "loaded": True,
            },
            {
                "name": "orcheo-plugin-bar",
                "enabled": False,
                "loaded": True,
            },
            {
                "name": "orcheo-plugin-baz",
                "enabled": True,
                "loaded": False,
            },
        ],
    )

    missing = plugin_inventory.missing_required_plugins(
        [
            "orcheo-plugin-foo",
            "orcheo-plugin-bar",
            " orcheo-plugin-baz ",
            "",
            "orcheo-plugin-bar",
        ]
    )
    assert missing == ["orcheo-plugin-bar", "orcheo-plugin-baz"]



"""Tests for ChatKit server helpers that parse widget payloads and route actions."""

from __future__ import annotations
import builtins
from collections.abc import Iterator, Mapping
from datetime import UTC, datetime
from types import SimpleNamespace
from typing import Any
from unittest.mock import AsyncMock, Mock
from uuid import uuid4
import pytest
from chatkit.errors import CustomStreamError
from chatkit.store import Store
from chatkit.types import (
    AssistantMessageContent,
    AssistantMessageItem,
    NoticeEvent,
    ThreadItemDoneEvent,
    ThreadMetadata,
    WidgetItem,
)
from dynaconf import Dynaconf
from orcheo_backend.app.chatkit import server as server_module
from orcheo_backend.app.chatkit.context import ChatKitRequestContext
from orcheo_backend.app.chatkit.telemetry import chatkit_telemetry
from orcheo_backend.app.repository import (
    WorkflowNotFoundError,
    WorkflowVersionNotFoundError,
)


class DummyStore(Store[ChatKitRequestContext]):
    """Minimal store implementation used by the server helpers."""

    def __init__(self) -> None:
        self.generated: list[str] = []
        self.saved: list[str] = []
        self.added: list[tuple[str, Any]] = []

    def generate_item_id(
        self,
        item_type: str,
        thread: ThreadMetadata,
        context: ChatKitRequestContext,
    ) -> str:
        identifier = f"{item_type}-{len(self.generated) + 1}"
        self.generated.append(identifier)
        return identifier

    async def load_thread(self, _, __):
        raise NotImplementedError

    async def save_thread(self, thread_id: str, context: ChatKitRequestContext) -> None:
        self.saved.append(thread_id)

    async def load_thread_items(self, *args, **kwargs):
        raise NotImplementedError

    async def save_attachment(self, *args, **kwargs):
        raise NotImplementedError

    async def load_attachment(self, *args, **kwargs):
        raise NotImplementedError

    async def delete_attachment(self, *args, **kwargs):
        raise NotImplementedError

    async def load_threads(self, *args, **kwargs):
        raise NotImplementedError

    async def add_thread_item(
        self, thread_id: str, item: Any, context: ChatKitRequestContext
    ) -> None:
        self.added.append((thread_id, item))

    async def save_item(self, *args, **kwargs):
        raise NotImplementedError

    async def load_item(self, *args, **kwargs):
        raise NotImplementedError

    async def delete_thread(self, *args, **kwargs):
        raise NotImplementedError

    async def delete_thread_item(self, *args, **kwargs):
        raise NotImplementedError


@pytest.fixture(autouse=True)
def reset_chatkit_metrics() -> None:
    """Reset chatkit telemetry between tests."""
    chatkit_telemetry.reset()
    yield
    chatkit_telemetry.reset()


@pytest.fixture(autouse=True)
def reset_widget_policy_state() -> Iterator[None]:
    """Keep widget policy globals consistent between tests."""
    original_widget_types = set(server_module._WIDGET_TYPES)
    original_action_types = set(server_module._ALLOWED_WIDGET_ACTION_TYPES)
    try:
        yield
    finally:
        server_module._WIDGET_TYPES.clear()
        server_module._WIDGET_TYPES.update(original_widget_types)
        server_module._ALLOWED_WIDGET_ACTION_TYPES.clear()
        server_module._ALLOWED_WIDGET_ACTION_TYPES.update(original_action_types)


def create_server() -> tuple[server_module.OrcheoChatKitServer, DummyStore]:
    store = DummyStore()
    repository = Mock()
    server = server_module.OrcheoChatKitServer(
        store=store,
        repository=repository,
        vault_provider=lambda: None,
    )
    return server, store


def test_coerce_config_set_handles_none_and_unknown_type() -> None:
    default = {"Card", "ListView"}

    assert server_module._coerce_config_set(None, default) == default
    assert server_module._coerce_config_set(123, default) == default


def test_coerce_config_set_parses_strings_and_iterables() -> None:
    default = {"Card", "ListView"}

    assert server_module._coerce_config_set(" custom , ListView ", default) == {
        "custom",
        "ListView",
    }
    assert server_module._coerce_config_set(" , ", default) == default
    assert server_module._coerce_config_set(["Card", "  ", "ListView"], default) == {
        "Card",
        "ListView",
    }
    assert server_module._coerce_config_set(("Widget", " "), default) == {"Widget"}
    assert server_module._coerce_config_set(frozenset({"Card", "Custom"}), default) == {
        "Card",
        "Custom",
    }


def test_action_type_for_logging_handles_mapping_and_attributes() -> None:
    mapping = {"type": "map"}
    assert server_module._action_type_for_logging(mapping) == "map"

    class ActionObject:
        type = "object"

    assert server_module._action_type_for_logging(ActionObject()) == "object"


def test_candidate_type_reads_from_mapping_and_attributes() -> None:
    assert server_module._candidate_type({"type": "Card"}) == "Card"

    class Candidate:
        type = "ListView"

    assert server_module._candidate_type(Candidate()) == "ListView"


def test_content_text_extracts_from_strings_and_lists() -> None:
    assert server_module._content_text("plain") == "plain"

    mixed = [
        {"text": "first"},
        SimpleNamespace(text="second"),
        {"content": "ignored"},
    ]
    assert server_module._content_text(mixed) == "first"
    assert server_module._content_text([SimpleNamespace(nontext=1)]) is None


def test_content_text_uses_attribute_fallback() -> None:
    entries = [
        {"content": "ignored"},
        SimpleNamespace(text="attr text"),
    ]
    assert server_module._content_text(entries) == "attr text"


def test_candidate_from_content_handles_invalid_json() -> None:
    assert server_module._candidate_from_content("not-json", None) is None


def test_candidate_from_content_rejects_non_widget_types() -> None:
    assert server_module._candidate_from_content('{"type": "Unknown"}', "copy") is None


def test_candidate_from_content_builds_widget_candidate() -> None:
    candidate = server_module._candidate_from_content(
        '{"type": "Card"}', copy_text="copy"
    )
    assert candidate is not None
    assert candidate.payload["type"] == "Card"
    assert candidate.copy_text == "copy"


def test_candidate_from_content_rejects_unknown_types() -> None:
    assert server_module._candidate_from_content('{"type": "Unknown"}', None) is None
    assert server_module._candidate_from_content("not json", None) is None


def test_candidate_from_artifact_returns_structured_content() -> None:
    artifact = {
        "structured_content": {"type": "ListView"},
        "copy_text": "copy",
    }
    candidate = server_module._candidate_from_artifact(artifact)
    assert candidate is not None
    assert candidate.payload["type"] == "ListView"
    assert candidate.copy_text == "copy"


def test_extract_widget_candidate_prefers_artifact() -> None:
    message = {
        "artifact": {
            "structured_content": {"type": "Card"},
            "copy_text": "text",
        },
        "content": '{"type": "ListView"}',
    }
    candidate = server_module._extract_widget_candidate(message)
    assert candidate is not None
    assert candidate.payload["type"] == "Card"
    assert candidate.copy_text == "text"


def test_extract_widget_candidate_falls_back_to_content() -> None:
    message = {"content": '{"type": "Card"}'}
    candidate = server_module._extract_widget_candidate(message)
    assert candidate is not None
    assert candidate.payload["type"] == "Card"


def test_extract_widget_candidate_accepts_artifact_without_type() -> None:
    message = {
        "artifact": {
            "structured_content": {},
            "copy_text": "no-type",
        }
    }
    candidate = server_module._extract_widget_candidate(message)
    assert candidate is not None
    assert candidate.payload == {}
    assert candidate.copy_text == "no-type"


def test_validate_widget_root_rejects_missing_type() -> None:
    with pytest.raises(server_module._WidgetHydrationError):
        server_module._validate_widget_root({"content": "no type"})


def test_validate_widget_root_rejects_large_payload() -> None:
    payload = {"type": "Card", "title": "x" * 100_000}
    with pytest.raises(server_module._WidgetHydrationError) as excinfo:
        server_module._validate_widget_root(payload)
    assert excinfo.value.reason == "too_large"


def test_notice_for_widget_error_variants() -> None:
    large_error = server_module._WidgetHydrationError(
        "too_large", detail="big", size_bytes=60_000
    )
    result = server_module._notice_for_widget_error(large_error)
    assert "too large" in result.message

    missing_error = server_module._WidgetHydrationError("invalid_widget")
    notice = server_module._notice_for_widget_error(missing_error)
    assert (
        notice.message == "The workflow returned a widget that could not be rendered."
    )


@pytest.mark.asyncio
async def test_hydrate_widget_items_returns_widget_and_notices() -> None:
    server, _store = create_server()
    thread = ThreadMetadata(
        id="thread",
        created_at=datetime.now(UTC),
        metadata={"workflow_id": "workflow"},
    )
    context: ChatKitRequestContext = {}
    message = {"type": "tool", "content": '{"type": "Card"}'}
    widget_items, notices = await server._hydrate_widget_items(
        thread, {"_messages": [message]}, context
    )
    assert len(widget_items) == 1
    assert not notices

    # Invalid payload should produce a notice
    invalid_message = {
        "type": "tool",
        "artifact": {"structured_content": {"foo": "bar"}},
    }
    widget_items, notices = await server._hydrate_widget_items(
        thread, {"_messages": [invalid_message]}, context
    )
    assert widget_items == []
    assert notices


@pytest.mark.asyncio
async def test_hydrate_widget_items_skips_non_tool_messages() -> None:
    server, _store = create_server()
    thread = ThreadMetadata(
        id="thread",
        created_at=datetime.now(UTC),
        metadata={"workflow_id": "workflow"},
    )
    context: ChatKitRequestContext = {}
    widget_items, notices = await server._hydrate_widget_items(
        thread, {"_messages": [{"type": "not_tool"}]}, context
    )
    assert widget_items == []
    assert notices == []


@pytest.mark.asyncio
async def test_hydrate_widget_items_records_metrics_on_error() -> None:
    server, _store = create_server()
    thread = ThreadMetadata(
        id="thread-metrics",
        created_at=datetime.now(UTC),
        metadata={"workflow_id": "workflow"},
    )
    context: ChatKitRequestContext = {}

    widget_items, notices = await server._hydrate_widget_items(
        thread,
        {
            "_messages": [
                {"type": "tool", "artifact": {"structured_content": {"foo": "bar"}}}
            ]
        },
        context,
    )

    assert widget_items == []
    assert notices
    assert (
        chatkit_telemetry.metrics().get("widget.validation_error.invalid_widget") == 1
    )


def test_resolve_chatkit_sqlite_path_uses_dynaconf_and_mappings() -> None:
    dynaconf = Dynaconf(settings_files=[], load_dotenv=False)
    dynaconf.set("CHATKIT_SQLITE_PATH", "~/dynaconf.sqlite")
    dynaconf_path = server_module._resolve_chatkit_sqlite_path(dynaconf)
    assert dynaconf_path == server_module.Path("~/dynaconf.sqlite").expanduser()

    mapping_path = server_module._resolve_chatkit_sqlite_path(
        {"CHATKIT_SQLITE_PATH": "~/map.sqlite"}
    )
    assert mapping_path == server_module.Path("~/map.sqlite").expanduser()

    class ConfigObject:
        chatkit_sqlite_path = "~/attr.sqlite"

    object_path = server_module._resolve_chatkit_sqlite_path(ConfigObject())
    assert object_path == server_module.Path("~/attr.sqlite").expanduser()

    default_path = server_module._resolve_chatkit_sqlite_path(object())
    assert default_path == server_module.Path("~/.orcheo/chatkit.sqlite").expanduser()


def test_workflow_id_from_thread_reads_metadata() -> None:
    thread = ThreadMetadata(
        id="thread",
        created_at=datetime.now(UTC),
        metadata={"workflow_id": "id"},
    )
    assert server_module._workflow_id_from_thread(thread) == "id"

    thread.metadata = {}
    assert server_module._workflow_id_from_thread(thread) is None


def test_is_tool_message_recognises_tool_types() -> None:
    assert server_module._is_tool_message({"type": "tool"})

    class ToolLike:
        type = "tool"

    assert not server_module._is_tool_message(ToolLike())
    assert not server_module._is_tool_message({"type": "other"})


def test_is_supported_action_type_returns_notice_and_metrics() -> None:
    server, _store = create_server()
    thread = ThreadMetadata(
        id="thread-action",
        created_at=datetime.now(UTC),
        metadata={"workflow_id": "wf"},
    )

    result = server._is_supported_action_type(thread, {"type": "noop"})

    assert not result.allowed
    assert result.notice is not None
    assert result.reason == "unsupported_widget_action"
    metrics = chatkit_telemetry.metrics()
    assert metrics.get("widget_action.unsupported.noop") == 1


def test_is_supported_action_type_allows_submit() -> None:
    server, _store = create_server()
    thread = ThreadMetadata(
        id="thread-allowed",
        created_at=datetime.now(UTC),
        metadata={"workflow_id": "wf"},
    )

    result = server._is_supported_action_type(thread, {"type": "submit"})

    assert result.allowed
    assert result.notice is None
    assert chatkit_telemetry.metrics() == {}


@pytest.mark.asyncio
async def test_action_skips_processing_when_action_blocked(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    server, _store = create_server()
    thread = ThreadMetadata(
        id="thread-blocked",
        created_at=datetime.now(UTC),
        metadata={"workflow_id": "wf"},
    )

    monkeypatch.setattr(server, "_ensure_workflow_metadata", lambda *_: None)
    monkeypatch.setattr(server, "_require_workflow_id", lambda *_: uuid4())
    blocked = server_module._ActionValidationResult(
        allowed=False,
        notice=None,
        reason="unsupported_widget_action",
        action_type="noop",
    )
    monkeypatch.setattr(server, "_is_supported_action_type", lambda *_: blocked)

    history = AsyncMock()
    monkeypatch.setattr(server, "_history", history)
    run_workflow = AsyncMock()
    monkeypatch.setattr(server, "_run_workflow", run_workflow)

    events: list[Any] = []
    async for event in server.action(thread, {"type": "noop"}, None, {}):
        events.append(event)

    assert events == []
    history.assert_not_awaited()
    run_workflow.assert_not_awaited()


def test_refresh_widget_policy_respects_configuration() -> None:
    original_widget_types = set(server_module._WIDGET_TYPES)
    original_action_types = set(server_module._ALLOWED_WIDGET_ACTION_TYPES)
    settings = Dynaconf(settings_files=[], load_dotenv=False)
    settings.set("CHATKIT_WIDGET_TYPES", ["Custom"])
    settings.set("CHATKIT_WIDGET_ACTION_TYPES", ["tap"])

    server_module._refresh_widget_policy(settings)

    try:
        assert server_module._WIDGET_TYPES == {"Custom"}
        assert server_module._ALLOWED_WIDGET_ACTION_TYPES == {"tap"}
    finally:
        reset_settings = Dynaconf(settings_files=[], load_dotenv=False)
        reset_settings.set("CHATKIT_WIDGET_TYPES", list(original_widget_types))
        reset_settings.set("CHATKIT_WIDGET_ACTION_TYPES", list(original_action_types))
        server_module._refresh_widget_policy(reset_settings)


def test_refresh_widget_policy_respects_mapping_without_get(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    class MappingWithoutGet(Mapping[str, list[str]]):
        def __init__(self) -> None:
            self._data = {
                "CHATKIT_WIDGET_TYPES": ["MappedCard"],
                "CHATKIT_WIDGET_ACTION_TYPES": ["mapped"],
            }

        def __getitem__(self, key: str) -> list[str]:
            return self._data[key]

        def __iter__(self):
            return iter(self._data)

        def __len__(self) -> int:
            return len(self._data)

    mapping = MappingWithoutGet()

    original_hasattr = builtins.hasattr

    def fake_hasattr(obj: object, name: str) -> bool:  # type: ignore[override]
        if obj is mapping and name == "get":
            return False
        return original_hasattr(obj, name)

    monkeypatch.setattr(builtins, "hasattr", fake_hasattr)

    server_module._refresh_widget_policy(mapping)

    assert server_module._WIDGET_TYPES == {"MappedCard"}
    assert server_module._ALLOWED_WIDGET_ACTION_TYPES == {"mapped"}


def test_refresh_widget_policy_respects_attribute_config() -> None:
    class AttributeConfig:
        chatkit_widget_types = ["AttributeCard"]
        chatkit_widget_action_types = ["tap"]

    server_module._refresh_widget_policy(AttributeConfig())

    assert server_module._WIDGET_TYPES == {"AttributeCard"}
    assert server_module._ALLOWED_WIDGET_ACTION_TYPES == {"tap"}


@pytest.mark.asyncio
async def test_action_handles_workflow_not_found(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    server, store = create_server()
    thread = ThreadMetadata(
        id="thread-not-found",
        created_at=datetime.now(UTC),
        metadata={"workflow_id": "wf"},
    )
    monkeypatch.setattr(server, "_ensure_workflow_metadata", lambda *_: None)
    monkeypatch.setattr(server, "_require_workflow_id", lambda *_: uuid4())
    monkeypatch.setattr(server, "_history", AsyncMock(return_value=[]))
    monkeypatch.setattr(
        server_module,
        "build_action_inputs_payload",
        lambda *_: {"payload": "value"},
    )
    mock_failure = AsyncMock(side_effect=WorkflowNotFoundError("missing"))
    monkeypatch.setattr(server, "_run_workflow", mock_failure)
    log_mock = Mock()
    server._log_action_failure = log_mock

    with pytest.raises(CustomStreamError):
        async for _ in server.action(thread, {"type": "submit"}, None, {}):
            ...
    assert log_mock.called


@pytest.mark.asyncio
async def test_action_handles_version_not_found(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    server, store = create_server()
    thread = ThreadMetadata(
        id="thread-version",
        created_at=datetime.now(UTC),
        metadata={"workflow_id": "wf"},
    )
    monkeypatch.setattr(server, "_ensure_workflow_metadata", lambda *_: None)
    monkeypatch.setattr(server, "_require_workflow_id", lambda *_: uuid4())
    monkeypatch.setattr(server, "_history", AsyncMock(return_value=[]))
    monkeypatch.setattr(
        server_module,
        "build_action_inputs_payload",
        lambda *_: {"payload": "value"},
    )
    mock_failure = AsyncMock(side_effect=WorkflowVersionNotFoundError("version"))
    monkeypatch.setattr(server, "_run_workflow", mock_failure)
    server._log_action_failure = Mock()

    with pytest.raises(CustomStreamError):
        async for _ in server.action(thread, {"type": "submit"}, None, {}):
            ...
    assert server._log_action_failure.called


@pytest.mark.asyncio
async def test_action_logs_and_reraises_generic_exceptions(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    server, store = create_server()
    thread = ThreadMetadata(
        id="thread-error",
        created_at=datetime.now(UTC),
        metadata={"workflow_id": "wf"},
    )
    monkeypatch.setattr(server, "_ensure_workflow_metadata", lambda *_: None)
    monkeypatch.setattr(server, "_require_workflow_id", lambda *_: uuid4())
    monkeypatch.setattr(server, "_history", AsyncMock(return_value=[]))
    monkeypatch.setattr(
        server_module,
        "build_action_inputs_payload",
        lambda *_: {"payload": "value"},
    )
    mock_failure = AsyncMock(side_effect=RuntimeError("boom"))
    monkeypatch.setattr(server, "_run_workflow", mock_failure)
    log_mock = Mock()
    server._log_action_failure = log_mock

    with pytest.raises(RuntimeError):
        async for _ in server.action(thread, {"type": "submit"}, None, {}):
            ...
    assert log_mock.called


@pytest.mark.asyncio
async def test_action_streams_widgets_and_assistant(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    server, store = create_server()
    thread = ThreadMetadata(
        id="thread-stream",
        created_at=datetime.now(UTC),
        metadata={"workflow_id": "wf"},
    )
    context: ChatKitRequestContext = {"actor": "actor"}
    monkeypatch.setattr(server, "_ensure_workflow_metadata", lambda *_: None)
    monkeypatch.setattr(server, "_require_workflow_id", lambda *_: uuid4())
    monkeypatch.setattr(server, "_history", AsyncMock(return_value=[]))
    monkeypatch.setattr(
        server_module,
        "build_action_inputs_payload",
        lambda *_: {"payload": "value"},
    )

    async def fake_run_workflow(*args: Any, **kwargs: Any):
        return "reply", {}, None

    monkeypatch.setattr(server, "_run_workflow", fake_run_workflow)
    notice = NoticeEvent(level="info", message="note")
    widget_item = WidgetItem(
        id="widget",
        thread_id="thread-stream",
        created_at=datetime.now(UTC),
        widget={"type": "Card"},
    )

    async def fake_hydrate(*_: Any):
        return [widget_item], [notice]

    monkeypatch.setattr(server, "_hydrate_widget_items", fake_hydrate)
    assistant_item = AssistantMessageItem(
        id="assistant",
        thread_id=thread.id,
        created_at=datetime.now(UTC),
        content=[AssistantMessageContent(text="reply")],
    )
    monkeypatch.setattr(server, "_build_assistant_item", lambda *_: assistant_item)
    store.add_thread_item = AsyncMock()
    store.save_thread = AsyncMock()
    events: list[Any] = []
    async for event in server.action(thread, {"type": "submit"}, None, context):
        events.append(event)

    assert events[0] is notice
    assert isinstance(events[1], ThreadItemDoneEvent)
    assert events[1].item is widget_item
    assert isinstance(events[2], ThreadItemDoneEvent)
    store.add_thread_item.assert_any_await(thread.id, widget_item, context)
    store.add_thread_item.assert_any_await(thread.id, assistant_item, context)
    store.save_thread.assert_awaited_once_with(thread, context)



"""Shared fixtures for workflow repository backend tests."""

from __future__ import annotations
from collections.abc import AsyncIterator, Generator
from pathlib import Path
from unittest.mock import patch
import pytest
import pytest_asyncio
from orcheo_backend.app.repository import (
    InMemoryWorkflowRepository,
    SqliteWorkflowRepository,
    WorkflowRepository,
)


@pytest.fixture(autouse=True)
def mock_celery_enqueue() -> Generator[None, None, None]:
    """Disable Celery enqueue for all repository tests to avoid Redis hangs."""
    with patch(
        "orcheo_backend.app.repository_sqlite._triggers._enqueue_run_for_execution"
    ):
        yield


@pytest_asyncio.fixture(params=["memory", "sqlite"])
async def repository(
    request: pytest.FixtureRequest, tmp_path_factory: pytest.TempPathFactory
) -> AsyncIterator[WorkflowRepository]:
    """Yield a repository instance backed by either the in-memory or SQLite backend."""

    if request.param == "memory":
        repo: WorkflowRepository = InMemoryWorkflowRepository()
    else:
        db_root = Path(tmp_path_factory.mktemp("repo"))
        repo = SqliteWorkflowRepository(db_root / "workflows.sqlite")

    try:
        yield repo
    finally:
        await repo.reset()



"""Tests for the trigger registry module."""

from __future__ import annotations
from typing import Any
import pytest
from orcheo.triggers.registry import TriggerMetadata, TriggerRegistry


def _make_registry() -> TriggerRegistry:
    return TriggerRegistry()


def test_register_and_get() -> None:
    """register stores metadata and factory; get retrieves the registration."""
    registry = _make_registry()
    metadata = TriggerMetadata(id="cron", display_name="Cron", description="Schedule")

    def factory(**kwargs: Any) -> dict[str, Any]:
        return {"kind": "cron", **kwargs}

    registry.register(metadata, factory)
    registration = registry.get("cron")

    assert registration is not None
    assert registration.metadata is metadata
    assert registration.factory is factory


def test_get_unknown_returns_none() -> None:
    """get returns None for unknown trigger ids."""
    registry = _make_registry()
    assert registry.get("unknown") is None


def test_register_duplicate_raises() -> None:
    """Registering the same id twice raises ValueError (lines 36-37)."""
    registry = _make_registry()
    meta = TriggerMetadata(id="webhook", display_name="Webhook")

    def noop() -> None:
        pass

    registry.register(meta, noop)

    with pytest.raises(ValueError, match="webhook"):
        registry.register(meta, noop)


def test_list_metadata_empty() -> None:
    """list_metadata returns an empty list when no triggers are registered."""
    registry = _make_registry()
    assert registry.list_metadata() == []


def test_list_metadata_sorted() -> None:
    """list_metadata returns entries sorted by id case-insensitively (line 49)."""
    registry = _make_registry()

    def noop() -> None:
        pass

    registry.register(TriggerMetadata(id="webhook", display_name="Webhook"), noop)
    registry.register(TriggerMetadata(id="Cron", display_name="Cron"), noop)
    registry.register(TriggerMetadata(id="manual", display_name="Manual"), noop)

    ids = [m.id for m in registry.list_metadata()]
    assert ids == ["Cron", "manual", "webhook"]


def test_unregister_removes_registration() -> None:
    """unregister removes an existing trigger registration."""
    registry = _make_registry()

    def noop() -> None:
        pass

    registry.register(TriggerMetadata(id="cron", display_name="Cron"), noop)
    registry.unregister("cron")
    assert registry.get("cron") is None


def test_unregister_unknown_is_noop() -> None:
    """unregister silently ignores unknown ids."""
    registry = _make_registry()
    registry.unregister("does-not-exist")  # should not raise



"""Application-level configuration models."""

from __future__ import annotations
from pathlib import Path
from typing import cast
from pydantic import BaseModel, Field, field_validator, model_validator
from orcheo.config.chatkit_rate_limit_settings import ChatKitRateLimitSettings
from orcheo.config.defaults import _DEFAULTS
from orcheo.config.types import (
    ChatKitBackend,
    CheckpointBackend,
    GraphStoreBackend,
    RepositoryBackend,
)
from orcheo.config.vault_settings import VaultSettings


class AppSettings(BaseModel):
    """Validated application runtime settings."""

    checkpoint_backend: CheckpointBackend = Field(
        default=cast(CheckpointBackend, _DEFAULTS["CHECKPOINT_BACKEND"])
    )
    sqlite_path: str = Field(default=cast(str, _DEFAULTS["SQLITE_PATH"]))
    graph_store_backend: GraphStoreBackend = Field(
        default=cast(GraphStoreBackend, _DEFAULTS["GRAPH_STORE_BACKEND"])
    )
    graph_store_sqlite_path: str = Field(
        default=cast(str, _DEFAULTS["GRAPH_STORE_SQLITE_PATH"])
    )
    repository_backend: RepositoryBackend = Field(
        default=cast(RepositoryBackend, _DEFAULTS["REPOSITORY_BACKEND"])
    )
    repository_sqlite_path: str = Field(
        default=cast(str, _DEFAULTS["REPOSITORY_SQLITE_PATH"])
    )
    chatkit_backend: ChatKitBackend = Field(
        default=cast(ChatKitBackend, _DEFAULTS["CHATKIT_BACKEND"])
    )
    chatkit_sqlite_path: str = Field(
        default=cast(str, _DEFAULTS["CHATKIT_SQLITE_PATH"])
    )
    chatkit_storage_path: str = Field(
        default=cast(str, _DEFAULTS["CHATKIT_STORAGE_PATH"])
    )
    chatkit_public_base_url: str | None = None
    chatkit_max_upload_size_bytes: int = Field(
        default=cast(int, _DEFAULTS["CHATKIT_MAX_UPLOAD_SIZE_BYTES"]), gt=0
    )
    chatkit_retention_days: int = Field(
        default=cast(int, _DEFAULTS["CHATKIT_RETENTION_DAYS"]), gt=0
    )
    chatkit_rate_limits: ChatKitRateLimitSettings = Field(
        default_factory=ChatKitRateLimitSettings
    )
    chatkit_widget_types: set[str] = Field(
        default_factory=lambda: set(cast(list[str], _DEFAULTS["CHATKIT_WIDGET_TYPES"]))
    )
    chatkit_widget_action_types: set[str] = Field(
        default_factory=lambda: set(
            cast(list[str], _DEFAULTS["CHATKIT_WIDGET_ACTION_TYPES"])
        )
    )
    postgres_dsn: str | None = None
    postgres_pool_min_size: int = Field(
        default=cast(int, _DEFAULTS["POSTGRES_POOL_MIN_SIZE"]), ge=1
    )
    postgres_pool_max_size: int = Field(
        default=cast(int, _DEFAULTS["POSTGRES_POOL_MAX_SIZE"]), ge=1
    )
    postgres_pool_timeout: float = Field(
        default=cast(float, _DEFAULTS["POSTGRES_POOL_TIMEOUT"]), gt=0.0
    )
    postgres_pool_max_idle: float = Field(
        default=cast(float, _DEFAULTS["POSTGRES_POOL_MAX_IDLE"]), gt=0.0
    )
    host: str = Field(default=cast(str, _DEFAULTS["HOST"]))
    port: int = Field(default=cast(int, _DEFAULTS["PORT"]))
    vault: VaultSettings = Field(default_factory=VaultSettings)
    tracing_exporter: str = Field(default=cast(str, _DEFAULTS["TRACING_EXPORTER"]))
    tracing_endpoint: str | None = None
    tracing_service_name: str = Field(
        default=cast(str, _DEFAULTS["TRACING_SERVICE_NAME"])
    )
    tracing_sample_ratio: float = Field(
        default=cast(float, _DEFAULTS["TRACING_SAMPLE_RATIO"]),
        ge=0.0,
        le=1.0,
    )
    tracing_insecure: bool = Field(default=cast(bool, _DEFAULTS["TRACING_INSECURE"]))
    tracing_high_token_threshold: int = Field(
        default=cast(int, _DEFAULTS["TRACING_HIGH_TOKEN_THRESHOLD"]),
        ge=1,
    )
    tracing_preview_max_length: int = Field(
        default=cast(int, _DEFAULTS["TRACING_PREVIEW_MAX_LENGTH"]),
        ge=16,
    )

    @staticmethod
    def _coerce_widget_set(value: object, default_key: str) -> set[str]:
        defaults = set(cast(list[str], _DEFAULTS[default_key]))
        if value is None:
            return defaults
        if isinstance(value, str):
            parts = [part.strip() for part in value.split(",") if part.strip()]
            return set(parts) or defaults
        if isinstance(value, set | frozenset | list | tuple):
            coerced = {str(entry).strip() for entry in value if str(entry).strip()}
            return coerced or defaults
        return defaults

    @field_validator("checkpoint_backend", mode="before")
    @classmethod
    def _coerce_checkpoint_backend(cls, value: object) -> CheckpointBackend:
        candidate = (
            cast(str, value).lower()
            if value is not None
            else cast(str, _DEFAULTS["CHECKPOINT_BACKEND"])
        )
        if candidate not in {"sqlite", "postgres"}:
            msg = "ORCHEO_CHECKPOINT_BACKEND must be either 'sqlite' or 'postgres'."
            raise ValueError(msg)
        return cast(CheckpointBackend, candidate)

    @field_validator("repository_backend", mode="before")
    @classmethod
    def _coerce_repository_backend(cls, value: object) -> RepositoryBackend:
        candidate = (
            cast(str, value).lower()
            if value is not None
            else cast(str, _DEFAULTS["REPOSITORY_BACKEND"])
        )
        if candidate not in {"inmemory", "sqlite", "postgres"}:
            msg = (
                "ORCHEO_REPOSITORY_BACKEND must be 'inmemory', 'sqlite', or 'postgres'."
            )
            raise ValueError(msg)
        return cast(RepositoryBackend, candidate)

    @field_validator("graph_store_backend", mode="before")
    @classmethod
    def _coerce_graph_store_backend(cls, value: object) -> GraphStoreBackend:
        candidate = (
            cast(str, value).lower()
            if value is not None
            else cast(str, _DEFAULTS["GRAPH_STORE_BACKEND"])
        )
        if candidate not in {"sqlite", "postgres"}:
            msg = "ORCHEO_GRAPH_STORE_BACKEND must be either 'sqlite' or 'postgres'."
            raise ValueError(msg)
        return cast(GraphStoreBackend, candidate)

    @field_validator("chatkit_backend", mode="before")
    @classmethod
    def _coerce_chatkit_backend(cls, value: object) -> ChatKitBackend:
        candidate = (
            cast(str, value).lower()
            if value is not None
            else cast(str, _DEFAULTS["CHATKIT_BACKEND"])
        )
        if candidate not in {"sqlite", "postgres"}:
            msg = "ORCHEO_CHATKIT_BACKEND must be either 'sqlite' or 'postgres'."
            raise ValueError(msg)
        return cast(ChatKitBackend, candidate)

    @field_validator("sqlite_path", "host", mode="before")
    @classmethod
    def _coerce_str(cls, value: object) -> str:
        return str(value) if value is not None else ""

    @field_validator("repository_sqlite_path", mode="before")
    @classmethod
    def _coerce_repo_sqlite_path(cls, value: object) -> str:
        return str(value) if value is not None else ""

    @field_validator("graph_store_sqlite_path", mode="before")
    @classmethod
    def _coerce_graph_store_sqlite_path(cls, value: object) -> str:
        return str(value) if value is not None else ""

    @field_validator("chatkit_sqlite_path", mode="before")
    @classmethod
    def _coerce_chatkit_sqlite_path(cls, value: object) -> str:
        return str(value) if value is not None else ""

    @field_validator("chatkit_storage_path", mode="before")
    @classmethod
    def _coerce_chatkit_storage_path(cls, value: object) -> str:
        return str(value) if value is not None else ""

    @field_validator("chatkit_public_base_url", mode="before")
    @classmethod
    def _coerce_chatkit_public_base_url(cls, value: object) -> str | None:
        if value is None:
            return None
        candidate = str(value).strip()
        return candidate or None

    @field_validator("chatkit_max_upload_size_bytes", mode="before")
    @classmethod
    def _coerce_chatkit_max_upload_size_bytes(cls, value: object) -> int:
        candidate_obj = (
            value if value is not None else _DEFAULTS["CHATKIT_MAX_UPLOAD_SIZE_BYTES"]
        )
        if isinstance(candidate_obj, int | float):
            return int(candidate_obj)
        try:
            return int(str(candidate_obj))
        except (TypeError, ValueError) as exc:  # pragma: no cover - defensive
            msg = "ORCHEO_CHATKIT_MAX_UPLOAD_SIZE_BYTES must be an integer."
            raise ValueError(msg) from exc

    @field_validator("chatkit_retention_days", mode="before")
    @classmethod
    def _coerce_chatkit_retention(cls, value: object) -> int:
        candidate_obj = (
            value if value is not None else _DEFAULTS["CHATKIT_RETENTION_DAYS"]
        )
        if isinstance(candidate_obj, int):
            return candidate_obj
        try:
            return int(str(candidate_obj))
        except (TypeError, ValueError) as exc:  # pragma: no cover - defensive
            msg = "ORCHEO_CHATKIT_RETENTION_DAYS must be an integer."
            raise ValueError(msg) from exc

    @field_validator("chatkit_widget_types", mode="before")
    @classmethod
    def _coerce_widget_types(cls, value: object) -> set[str]:
        return cls._coerce_widget_set(value, "CHATKIT_WIDGET_TYPES")

    @field_validator("chatkit_widget_action_types", mode="before")
    @classmethod
    def _coerce_widget_action_types(cls, value: object) -> set[str]:
        return cls._coerce_widget_set(value, "CHATKIT_WIDGET_ACTION_TYPES")

    @staticmethod
    def _coerce_postgres_pool_int(value: object, default_key: str) -> int:
        if value is None:
            return cast(int, _DEFAULTS[default_key])
        if isinstance(value, int):
            return value
        try:
            return int(str(value))
        except (TypeError, ValueError) as exc:  # pragma: no cover - defensive
            msg = "PostgreSQL pool size must be a positive integer."
            raise ValueError(msg) from exc

    @staticmethod
    def _coerce_postgres_pool_float(value: object, default_key: str) -> float:
        if value is None:
            return cast(float, _DEFAULTS[default_key])
        if isinstance(value, int | float):
            return float(value)
        try:
            return float(str(value))
        except (TypeError, ValueError) as exc:  # pragma: no cover - defensive
            msg = "PostgreSQL pool timeout must be a positive number."
            raise ValueError(msg) from exc

    @field_validator("postgres_pool_min_size", mode="before")
    @classmethod
    def _coerce_postgres_pool_min_size(cls, value: object) -> int:
        return cls._coerce_postgres_pool_int(value, "POSTGRES_POOL_MIN_SIZE")

    @field_validator("postgres_pool_max_size", mode="before")
    @classmethod
    def _coerce_postgres_pool_max_size(cls, value: object) -> int:
        return cls._coerce_postgres_pool_int(value, "POSTGRES_POOL_MAX_SIZE")

    @field_validator("postgres_pool_timeout", mode="before")
    @classmethod
    def _coerce_postgres_pool_timeout(cls, value: object) -> float:
        return cls._coerce_postgres_pool_float(value, "POSTGRES_POOL_TIMEOUT")

    @field_validator("postgres_pool_max_idle", mode="before")
    @classmethod
    def _coerce_postgres_pool_max_idle(cls, value: object) -> float:
        return cls._coerce_postgres_pool_float(value, "POSTGRES_POOL_MAX_IDLE")

    @field_validator("port", mode="before")
    @classmethod
    def _parse_port(cls, value: object) -> int:
        candidate_obj = value if value is not None else _DEFAULTS["PORT"]
        candidate: int | str
        if isinstance(candidate_obj, int | str):
            candidate = candidate_obj
        else:
            candidate = str(candidate_obj)
        try:
            return int(candidate)
        except (TypeError, ValueError) as exc:  # pragma: no cover - defensive
            raise ValueError("ORCHEO_PORT must be an integer.") from exc

    @field_validator("tracing_exporter", mode="before")
    @classmethod
    def _coerce_tracing_exporter(cls, value: object) -> str:
        candidate_source = value if value is not None else _DEFAULTS["TRACING_EXPORTER"]
        candidate = str(candidate_source).lower()
        allowed = {"none", "otlp", "console"}
        if candidate not in allowed:
            msg = "ORCHEO_TRACING_EXPORTER must be one of 'none', 'otlp', or 'console'."
            raise ValueError(msg)
        return candidate

    @field_validator("tracing_endpoint", mode="before")
    @classmethod
    def _coerce_tracing_endpoint(cls, value: object) -> str | None:
        if value is None:
            return None
        return str(value)

    @field_validator("tracing_service_name", mode="before")
    @classmethod
    def _coerce_tracing_service_name(cls, value: object) -> str:
        candidate = value if value is not None else _DEFAULTS["TRACING_SERVICE_NAME"]
        return str(candidate)

    @field_validator("tracing_sample_ratio", mode="before")
    @classmethod
    def _coerce_tracing_sample_ratio(cls, value: object) -> float:
        candidate_obj = (
            value if value is not None else _DEFAULTS["TRACING_SAMPLE_RATIO"]
        )
        try:
            if isinstance(candidate_obj, int | float | str):
                candidate = float(candidate_obj)
            else:
                candidate = float(str(candidate_obj))
        except (TypeError, ValueError) as exc:  # pragma: no cover - defensive
            msg = "ORCHEO_TRACING_SAMPLE_RATIO must be a float between 0 and 1."
            raise ValueError(msg) from exc
        if not 0.0 <= candidate <= 1.0:
            msg = "ORCHEO_TRACING_SAMPLE_RATIO must be between 0 and 1."
            raise ValueError(msg)
        return candidate

    @field_validator("tracing_insecure", mode="before")
    @classmethod
    def _coerce_tracing_insecure(cls, value: object) -> bool:
        candidate_obj = value if value is not None else _DEFAULTS["TRACING_INSECURE"]
        if isinstance(candidate_obj, bool):
            return candidate_obj
        if isinstance(candidate_obj, str):
            lowered = candidate_obj.strip().lower()
            if lowered in {"1", "true", "yes", "on"}:
                return True
            if lowered in {"0", "false", "no", "off"}:
                return False
        return bool(candidate_obj)

    @field_validator("tracing_high_token_threshold", mode="before")
    @classmethod
    def _coerce_tracing_high_token_threshold(cls, value: object) -> int:
        candidate_obj = (
            value if value is not None else _DEFAULTS["TRACING_HIGH_TOKEN_THRESHOLD"]
        )
        if isinstance(candidate_obj, bool):  # pragma: no cover - defensive
            return int(candidate_obj)
        if isinstance(candidate_obj, int | float):
            return int(candidate_obj)
        try:
            return int(str(candidate_obj))
        except (TypeError, ValueError) as exc:  # pragma: no cover - defensive
            msg = "ORCHEO_TRACING_HIGH_TOKEN_THRESHOLD must be a positive integer."
            raise ValueError(msg) from exc

    @field_validator("tracing_preview_max_length", mode="before")
    @classmethod
    def _coerce_tracing_preview_max_length(cls, value: object) -> int:
        candidate_obj = (
            value if value is not None else _DEFAULTS["TRACING_PREVIEW_MAX_LENGTH"]
        )
        if isinstance(candidate_obj, bool):  # pragma: no cover - defensive
            return int(candidate_obj)
        if isinstance(candidate_obj, int | float):
            return int(candidate_obj)
        try:
            return int(str(candidate_obj))
        except (TypeError, ValueError) as exc:  # pragma: no cover - defensive
            msg = "ORCHEO_TRACING_PREVIEW_MAX_LENGTH must be a positive integer."
            raise ValueError(msg) from exc

    def _validate_postgres_dsn(self) -> None:
        uses_postgres = {
            self.checkpoint_backend,
            self.graph_store_backend,
            self.repository_backend,
            self.chatkit_backend,
            self.vault.backend,
        }
        if "postgres" in uses_postgres:
            if not self.postgres_dsn:
                msg = "ORCHEO_POSTGRES_DSN must be set when using the postgres backend."
                raise ValueError(msg)
            self.postgres_dsn = str(self.postgres_dsn)
            return
        self.postgres_dsn = None

    def _apply_runtime_defaults(self) -> None:
        self.sqlite_path = self.sqlite_path or cast(str, _DEFAULTS["SQLITE_PATH"])
        self.graph_store_sqlite_path = self.graph_store_sqlite_path or cast(
            str, _DEFAULTS["GRAPH_STORE_SQLITE_PATH"]
        )
        self.repository_sqlite_path = self.repository_sqlite_path or cast(
            str, _DEFAULTS["REPOSITORY_SQLITE_PATH"]
        )
        self.chatkit_sqlite_path = self.chatkit_sqlite_path or cast(
            str, _DEFAULTS["CHATKIT_SQLITE_PATH"]
        )
        self.chatkit_storage_path = self.chatkit_storage_path or cast(
            str, _DEFAULTS["CHATKIT_STORAGE_PATH"]
        )
        if self.chatkit_retention_days <= 0:  # pragma: no cover - defensive
            self.chatkit_retention_days = cast(int, _DEFAULTS["CHATKIT_RETENTION_DAYS"])
        if not isinstance(self.chatkit_rate_limits, ChatKitRateLimitSettings):
            self.chatkit_rate_limits = ChatKitRateLimitSettings()
        self.host = self.host or cast(str, _DEFAULTS["HOST"])
        if not self.tracing_exporter:
            self.tracing_exporter = cast(str, _DEFAULTS["TRACING_EXPORTER"])
        if not self.tracing_service_name:
            self.tracing_service_name = cast(str, _DEFAULTS["TRACING_SERVICE_NAME"])
        if self.tracing_high_token_threshold <= 0:
            self.tracing_high_token_threshold = cast(
                int, _DEFAULTS["TRACING_HIGH_TOKEN_THRESHOLD"]
            )
        if self.tracing_preview_max_length <= 0:
            self.tracing_preview_max_length = cast(
                int, _DEFAULTS["TRACING_PREVIEW_MAX_LENGTH"]
            )

    def _validate_graph_store_path(self) -> None:
        if self.graph_store_backend != "sqlite":
            return
        if self.graph_store_sqlite_path.strip().startswith("~"):
            msg = (
                "ORCHEO_GRAPH_STORE_SQLITE_PATH must be an absolute path and "
                "must not start with '~'."
            )
            raise ValueError(msg)
        graph_store_path = Path(self.graph_store_sqlite_path)
        if not graph_store_path.is_absolute():
            msg = "ORCHEO_GRAPH_STORE_SQLITE_PATH must be an absolute filesystem path."
            raise ValueError(msg)
        self.graph_store_sqlite_path = str(graph_store_path.resolve(strict=False))

    @model_validator(mode="after")
    def _validate_postgres_requirements(self) -> AppSettings:
        self._validate_postgres_dsn()
        self._apply_runtime_defaults()
        self._validate_graph_store_path()
        return self


__all__ = ["AppSettings"]



"""Serialization helpers for the PostgreSQL ChatKit store."""

from __future__ import annotations
import json
from collections.abc import Mapping
from datetime import datetime
from typing import Any
from chatkit.types import Attachment, ThreadItem, ThreadMetadata
from pydantic import TypeAdapter
from orcheo_backend.app.chatkit_store_postgres.utils import (
    compact_json,
    ensure_datetime,
)


_THREAD_ADAPTER: TypeAdapter[ThreadMetadata] = TypeAdapter(ThreadMetadata)
_ITEM_ADAPTER: TypeAdapter[ThreadItem] = TypeAdapter(ThreadItem)
_ATTACHMENT_ADAPTER: TypeAdapter[Attachment] = TypeAdapter(Attachment)


def _coerce_json(value: Any) -> Any:
    if value is None:
        return None
    if isinstance(value, str):
        return json.loads(value)
    return value


def _coerce_datetime(value: Any) -> datetime:
    if isinstance(value, datetime):
        return ensure_datetime(value)
    if isinstance(value, str):
        return ensure_datetime(datetime.fromisoformat(value))
    raise TypeError("Unsupported datetime value")


def serialize_thread_status(thread: ThreadMetadata) -> str:
    """Serialize the thread status to a compact JSON string."""
    status_payload = (
        thread.status.model_dump(mode="json")
        if hasattr(thread.status, "model_dump")
        else thread.status
    )
    return compact_json(status_payload)


def thread_from_row(row: Mapping[str, Any]) -> ThreadMetadata:
    """Convert a SQL row into a ``ThreadMetadata`` instance."""
    data = {
        "id": row["id"],
        "title": row.get("title"),
        "created_at": _coerce_datetime(row["created_at"]),
        "status": _coerce_json(row["status_json"]),
        "metadata": _coerce_json(row["metadata_json"]),
    }
    return _THREAD_ADAPTER.validate_python(data)


def serialize_item(item: ThreadItem) -> str:
    """Serialize a ``ThreadItem`` to JSON."""
    payload = _ITEM_ADAPTER.dump_python(item, mode="json")
    return compact_json(payload)


def item_from_row(row: Mapping[str, Any]) -> ThreadItem:
    """Convert a SQL row into a ``ThreadItem``."""
    payload = _coerce_json(row["item_json"])
    if payload is None:
        payload = {}
    payload.setdefault("id", row["id"])
    payload.setdefault("thread_id", row["thread_id"])
    created_at = row.get("created_at")
    if created_at is not None:
        payload.setdefault("created_at", created_at)
    if isinstance(payload.get("created_at"), str):
        payload["created_at"] = datetime.fromisoformat(payload["created_at"])
    elif isinstance(payload.get("created_at"), datetime):
        payload["created_at"] = ensure_datetime(payload["created_at"])
    return _ITEM_ADAPTER.validate_python(payload)


def serialize_attachment(attachment: Attachment) -> str:
    """Serialize an attachment definition."""
    payload = _ATTACHMENT_ADAPTER.dump_python(attachment, mode="json")
    return compact_json(payload)


def attachment_from_details(details_json: Any) -> Attachment:
    """Deserialize attachment metadata."""
    payload = _coerce_json(details_json)
    return _ATTACHMENT_ADAPTER.validate_python(payload)


__all__ = [
    "attachment_from_details",
    "item_from_row",
    "serialize_attachment",
    "serialize_item",
    "serialize_thread_status",
    "thread_from_row",
]



"""Tests for skill service operations."""

from __future__ import annotations
from unittest.mock import MagicMock, patch
from orcheo.skills.models import SkillMetadata, SkillValidationResult
from orcheo_sdk.services.skills import (
    install_skill_data,
    list_skills_data,
    show_skill_data,
    uninstall_skill_data,
    validate_skill_data,
)


def test_list_skills_data_delegates_to_manager() -> None:
    """list_skills_data returns SkillManager().list_skills()."""
    expected = [{"name": "my-skill"}]
    mock_manager = MagicMock()
    mock_manager.list_skills.return_value = expected

    with patch("orcheo_sdk.services.skills.SkillManager", return_value=mock_manager):
        result = list_skills_data()

    mock_manager.list_skills.assert_called_once()
    assert result is expected


def test_show_skill_data_delegates_to_manager() -> None:
    """show_skill_data returns SkillManager().show_skill(name)."""
    expected = {"name": "my-skill", "description": "Desc."}
    mock_manager = MagicMock()
    mock_manager.show_skill.return_value = expected

    with patch("orcheo_sdk.services.skills.SkillManager", return_value=mock_manager):
        result = show_skill_data("my-skill")

    mock_manager.show_skill.assert_called_once_with("my-skill")
    assert result is expected


def test_install_skill_data_delegates_to_manager() -> None:
    """install_skill_data returns SkillManager().install(ref)."""
    expected = {"name": "my-skill"}
    mock_manager = MagicMock()
    mock_manager.install.return_value = expected

    with patch("orcheo_sdk.services.skills.SkillManager", return_value=mock_manager):
        result = install_skill_data("/path/to/skill")

    mock_manager.install.assert_called_once_with("/path/to/skill")
    assert result is expected


def test_uninstall_skill_data_delegates_to_manager() -> None:
    """uninstall_skill_data returns SkillManager().uninstall(name)."""
    expected = {"name": "my-skill", "status": "uninstalled"}
    mock_manager = MagicMock()
    mock_manager.uninstall.return_value = expected

    with patch("orcheo_sdk.services.skills.SkillManager", return_value=mock_manager):
        result = uninstall_skill_data("my-skill")

    mock_manager.uninstall.assert_called_once_with("my-skill")
    assert result is expected


def test_validate_skill_data_delegates_to_manager() -> None:
    """validate_skill_data returns SkillManager().validate(ref)."""
    expected = SkillValidationResult(
        valid=True,
        skill_metadata=SkillMetadata(name="my-skill", description="Desc."),
    )
    mock_manager = MagicMock()
    mock_manager.validate.return_value = expected

    with patch("orcheo_sdk.services.skills.SkillManager", return_value=mock_manager):
        result = validate_skill_data("/path/to/skill")

    mock_manager.validate.assert_called_once_with("/path/to/skill")
    assert result is expected



"""Keep Canvas template previews aligned with ingested workflow Mermaid."""

from __future__ import annotations
import re
import sys
import types
from pathlib import Path
import pytest
from langchain_core.runnables import RunnableConfig
from orcheo.graph.ingestion import ingest_langgraph_script
from orcheo.graph.state import State
from orcheo.nodes.base import TaskNode
from orcheo.nodes.listeners import ListenerNode


_REPO_ROOT = Path(__file__).resolve().parents[2]
_TEMPLATE_DIR = _REPO_ROOT / "apps/canvas/src/features/workflow/data/templates"
_ASSET_IMPORT_RE = re.compile(
    r'import\s+\w+\s+from\s+"\.\/assets\/([^\"]+)\/workflow\.py\?raw";'
)
_MERMAID_RE = re.compile(r"const\s+\w+_MERMAID\s*=\s*`(.*?)`;", re.S)


class _WeComListenerPluginNode(ListenerNode):
    platform: str = "wecom"
    bot_id: str = ""
    bot_secret: str = ""


class _LarkListenerPluginNode(ListenerNode):
    platform: str = "lark"
    app_id: str = ""
    app_secret: str = ""


class _WeComWsReplyNode(TaskNode):
    message: str = ""
    raw_event: str = ""
    subscription_id: str = ""

    async def run(self, state: State, config: RunnableConfig) -> dict[str, str | None]:
        del state, config
        return {}


def _template_paths() -> list[Path]:
    return sorted(
        path
        for path in _TEMPLATE_DIR.glob("*.ts")
        if _MERMAID_RE.search(path.read_text())
    )


@pytest.fixture()
def plugin_template_modules(monkeypatch: pytest.MonkeyPatch) -> None:
    """Install lightweight plugin stubs so plugin-backed templates can ingest."""

    wecom_module = types.ModuleType("orcheo_plugin_wecom_listener")
    wecom_module.WeComListenerPluginNode = _WeComListenerPluginNode
    wecom_module.WeComWsReplyNode = _WeComWsReplyNode
    lark_module = types.ModuleType("orcheo_plugin_lark_listener")
    lark_module.LarkListenerPluginNode = _LarkListenerPluginNode
    wechat_module = types.ModuleType("orcheo_plugin_wechat_listener")

    class _WechatListenerPluginNode(ListenerNode):
        platform: str = "wechat"
        account_id: str = ""
        bot_token: str = ""
        base_url: str = ""
        bot_identity_key: str = ""

    class _WechatReplyNode(TaskNode):
        account_id: str = ""
        bot_token: str = ""
        base_url: str = ""
        message: str = ""
        reply_target: dict[str, str] | str = {}
        raw_event: dict[str, str] | str = {}

        async def run(
            self, state: State, config: RunnableConfig
        ) -> dict[str, bool | str | None]:
            del state, config
            return {"sent": True}

    wechat_module.WechatListenerPluginNode = _WechatListenerPluginNode
    wechat_module.WechatReplyNode = _WechatReplyNode

    monkeypatch.setitem(sys.modules, "orcheo_plugin_wecom_listener", wecom_module)
    monkeypatch.setitem(sys.modules, "orcheo_plugin_lark_listener", lark_module)
    monkeypatch.setitem(sys.modules, "orcheo_plugin_wechat_listener", wechat_module)


@pytest.mark.parametrize(
    "template_path",
    _template_paths(),
    ids=lambda path: path.name,
)
def test_canvas_template_mermaid_matches_ingested_workflow_preview(
    template_path: Path,
    plugin_template_modules: None,
) -> None:
    """Template cards should render the same Mermaid preview as saved workflows."""

    template_source = template_path.read_text()
    asset_import_match = _ASSET_IMPORT_RE.search(template_source)
    mermaid_match = _MERMAID_RE.search(template_source)

    assert asset_import_match is not None
    assert mermaid_match is not None

    script_path = _TEMPLATE_DIR / "assets" / asset_import_match.group(1) / "workflow.py"
    ingested_mermaid = ingest_langgraph_script(script_path.read_text())["index"].get(
        "mermaid"
    )

    assert ingested_mermaid is not None
    assert mermaid_match.group(1).strip() == ingested_mermaid.strip()



"""Utility helpers for updating credential template fields."""

from __future__ import annotations
from collections.abc import Sequence
from orcheo.models import (
    CredentialIssuancePolicy,
    CredentialKind,
    CredentialScope,
    CredentialTemplate,
)


def update_template_simple_field(
    template: CredentialTemplate,
    attr: str,
    value: str | None,
    changes: dict[str, object],
) -> None:
    """Update a template scalar field while recording change metadata."""
    if value is None:
        return
    current = getattr(template, attr)
    if current == value:
        return
    changes[attr] = {"from": current, "to": value}
    setattr(template, attr, value)


def update_template_scopes(
    template: CredentialTemplate,
    scopes: Sequence[str] | None,
    changes: dict[str, object],
) -> None:
    """Update template scopes while tracking original values."""
    if scopes is None:
        return
    normalized = list(scopes)
    if normalized == template.scopes:
        return
    changes["scopes"] = {
        "from": list(template.scopes),
        "to": normalized,
    }
    template.scopes = normalized


def update_template_scope(
    template: CredentialTemplate,
    scope: CredentialScope | None,
    changes: dict[str, object],
) -> None:
    """Update the template scope and record the diff."""
    if scope is None or scope == template.scope:
        return
    changes["scope"] = {
        "from": template.scope.model_dump(),
        "to": scope.model_dump(),
    }
    template.scope = scope


def normalize_template_kind(
    kind: CredentialKind | str | None,
) -> CredentialKind | None:
    """Normalize user-provided template kind values to the enum."""
    if kind is None:
        return None
    if isinstance(kind, CredentialKind):
        return kind
    return CredentialKind(str(kind))


def update_template_kind(
    template: CredentialTemplate,
    kind: CredentialKind | str | None,
    changes: dict[str, object],
) -> None:
    """Update the template's credential kind when it changes."""
    new_kind = normalize_template_kind(kind)
    if new_kind is None or new_kind is template.kind:
        return
    changes["kind"] = {
        "from": template.kind.value,
        "to": new_kind.value,
    }
    template.kind = new_kind


def update_template_policy(
    template: CredentialTemplate,
    issuance_policy: CredentialIssuancePolicy | None,
    changes: dict[str, object],
) -> None:
    """Update the template issuance policy when a new value is provided."""
    if issuance_policy is None:
        return
    if issuance_policy.model_dump() == template.issuance_policy.model_dump():
        return
    changes["issuance_policy"] = {
        "from": template.issuance_policy.model_dump(),
        "to": issuance_policy.model_dump(),
    }
    template.issuance_policy = issuance_policy


__all__ = [
    "normalize_template_kind",
    "update_template_kind",
    "update_template_policy",
    "update_template_scope",
    "update_template_scopes",
    "update_template_simple_field",
]



"""Registry implementation for Orcheo agent tools."""

from collections.abc import Callable
from langchain_core.tools import BaseTool
from pydantic import BaseModel


class ToolMetadata(BaseModel):
    """Metadata for a tool in the agent.

    Attributes:
        name: Unique identifier for the tool
        description: Human-readable description of the tool's purpose
        category: Tool category, defaults to "general"
    """

    name: str
    """Unique identifier for the tool."""
    description: str
    """Human-readable description of the tool's purpose."""
    category: str = "general"
    """Tool category, defaults to "general"."""


class ToolRegistry:
    """Registry for managing agent tools and their metadata."""

    def __init__(self) -> None:
        """Initialize an empty tool registry."""
        self._tools: dict[str, BaseTool | Callable] = {}
        self._metadata: dict[str, ToolMetadata] = {}

    def register(
        self, metadata: ToolMetadata
    ) -> Callable[[Callable | BaseTool], Callable | BaseTool]:
        """Register a new tool with its metadata.

        Args:
            metadata: Tool metadata including name and description

        Returns:
            Decorator function that registers the tool or tool factory
        """

        def decorator(func: Callable | BaseTool) -> Callable | BaseTool:
            self._tools[metadata.name] = func
            self._metadata[metadata.name] = metadata
            return func

        return decorator

    def get_tool(self, name: str) -> BaseTool | Callable | None:
        """Get a tool by name.

        Args:
            name: Name of the tool to retrieve

        Returns:
            Tool instance or callable, or None if not found
        """
        return self._tools.get(name)

    def get_metadata(self, name: str) -> ToolMetadata | None:
        """Return metadata for the tool identified by ``name`` if available."""
        return self._metadata.get(name)

    def list_metadata(self) -> list[ToolMetadata]:
        """Return all registered tool metadata entries sorted by name."""
        return sorted(self._metadata.values(), key=lambda item: item.name.lower())

    def get_metadata_by_callable(self, obj: Callable) -> ToolMetadata | None:
        """Return metadata associated with a registered callable."""
        for name, registered in self._tools.items():
            if registered is obj:
                return self._metadata.get(name)
            if isinstance(registered, type) and isinstance(obj, registered):
                return self._metadata.get(name)
        return None

    def unregister(self, name: str) -> None:
        """Remove one registered tool and its metadata if present."""
        self._tools.pop(name, None)
        self._metadata.pop(name, None)


# Global registry instance
tool_registry = ToolRegistry()



"""Shared base class for credential vault implementations."""

from __future__ import annotations
import secrets
from orcheo.models import AesGcmCredentialCipher, CredentialCipher
from orcheo.vault.alerts import GovernanceAlertOperationsMixin
from orcheo.vault.credentials import CredentialOperationsMixin
from orcheo.vault.templates import TemplateOperationsMixin


class BaseCredentialVault(
    CredentialOperationsMixin,
    TemplateOperationsMixin,
    GovernanceAlertOperationsMixin,
):
    """Base helper that implements common credential vault workflows."""

    def __init__(self, *, cipher: CredentialCipher | None = None) -> None:
        """Initialize the vault with an encryption cipher."""
        self._cipher = cipher or AesGcmCredentialCipher(key=secrets.token_hex(32))

    @property
    def cipher(self) -> CredentialCipher:
        """Expose the credential cipher for services that need direct access."""
        return self._cipher


__all__ = ["BaseCredentialVault"]



"""Public entrypoints for LangGraph script ingestion."""

from __future__ import annotations
import sys as _sys_module
import threading as _threading_module
import time as _time_module
from contextlib import AbstractContextManager
from typing import Any
from RestrictedPython import compile_restricted as _restricted_compile_restricted
from orcheo.graph.ingestion.config import (
    DEFAULT_EXECUTION_TIMEOUT_SECONDS,
    DEFAULT_SCRIPT_SIZE_LIMIT,
    LANGGRAPH_SCRIPT_FORMAT,
)
from orcheo.graph.ingestion.exceptions import ScriptIngestionError
from orcheo.graph.ingestion.loader import _resolve_graph as _loader_resolve_graph
from orcheo.graph.ingestion.loader import load_graph_from_script
from orcheo.graph.ingestion.sandbox import (
    compile_langgraph_script as _sandbox_compile_langgraph_script,
)
from orcheo.graph.ingestion.sandbox import (
    execution_timeout as _sandbox_execution_timeout,
)
from orcheo.graph.ingestion.sandbox import (
    validate_script_size as _sandbox_validate_script_size,
)
from orcheo.graph.ingestion.summary import (
    _serialise_branch as _summary_serialise_branch,
)
from orcheo.graph.ingestion.summary import _unwrap_runnable as _summary_unwrap_runnable
from orcheo.graph.ingestion.summary import summarise_graph_index


_compile_langgraph_script = _sandbox_compile_langgraph_script
_resolve_graph = _loader_resolve_graph
_serialise_branch = _summary_serialise_branch
_unwrap_runnable = _summary_unwrap_runnable
_validate_script_size = _sandbox_validate_script_size
compile_restricted = _restricted_compile_restricted
sys = _sys_module
threading = _threading_module
time = _time_module


def ingest_langgraph_script(
    source: str,
    *,
    entrypoint: str | None = None,
    max_script_bytes: int | None = DEFAULT_SCRIPT_SIZE_LIMIT,
    execution_timeout_seconds: float | None = DEFAULT_EXECUTION_TIMEOUT_SECONDS,
) -> dict[str, Any]:
    """Return a workflow graph payload produced from a LangGraph Python script.

    The returned payload embeds the original script alongside a compact index
    containing derived metadata (for example, Mermaid output and cron trigger
    fields) while the original script is required to faithfully rebuild the
    graph during execution.
    """
    graph = load_graph_from_script(
        source,
        entrypoint=entrypoint,
        max_script_bytes=max_script_bytes,
        execution_timeout_seconds=execution_timeout_seconds,
    )
    index = summarise_graph_index(graph)
    return {
        "format": LANGGRAPH_SCRIPT_FORMAT,
        "source": source,
        "entrypoint": entrypoint,
        "index": index,
    }


def _execution_timeout(timeout_seconds: float | None) -> AbstractContextManager[None]:
    """Expose execution timeout that honours module-level monkeypatching."""
    return _sandbox_execution_timeout(
        timeout_seconds,
        sys_module=sys,
        threading_module=threading,
        time_module=time,
    )


__all__ = [
    "DEFAULT_EXECUTION_TIMEOUT_SECONDS",
    "DEFAULT_SCRIPT_SIZE_LIMIT",
    "LANGGRAPH_SCRIPT_FORMAT",
    "ScriptIngestionError",
    "compile_restricted",
    "_compile_langgraph_script",
    "_execution_timeout",
    "_resolve_graph",
    "_serialise_branch",
    "_unwrap_runnable",
    "_validate_script_size",
    "sys",
    "threading",
    "time",
    "ingest_langgraph_script",
    "load_graph_from_script",
    "summarise_graph_index",
]



"""Tests for plugin lifecycle CLI commands."""

from __future__ import annotations
import json
import shutil
import subprocess
from pathlib import Path
from typing import Any
from unittest.mock import MagicMock, patch
import pytest
import typer
from rich.console import Console
from typer.testing import CliRunner
from orcheo_sdk.cli import plugin as plugin_module
from orcheo_sdk.cli.main import app
from orcheo_sdk.cli.state import CLIState


FIXTURE_ROOT = Path(__file__).resolve().parents[1] / "plugin_fixtures"


def _copy_fixture(tmp_path: Path, fixture_name: str) -> Path:
    source = FIXTURE_ROOT / fixture_name
    destination = tmp_path / fixture_name
    shutil.copytree(source, destination)
    return destination


def _plugin_dir(env: dict[str, str]) -> Path:
    return Path(env["ORCHEO_PLUGIN_DIR"])


def _stack_env(base_env: dict[str, str], stack_dir: Path) -> dict[str, str]:
    env = dict(base_env)
    env["ORCHEO_STACK_DIR"] = str(stack_dir)
    return env


def _make_cli_state(*, human: bool) -> CLIState:
    return CLIState(
        settings=MagicMock(),
        client=MagicMock(),
        cache=MagicMock(),
        console=Console(record=True),
        human=human,
    )


def test_run_stack_subprocess_defaults_to_zero(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(
        "orcheo_sdk.cli.plugin.shutil.which", lambda _: "/usr/bin/docker"
    )

    def fake_run(
        command: list[str], *, check: bool, capture_output: bool, text: bool
    ) -> subprocess.CompletedProcess[str]:
        del check, capture_output, text
        return subprocess.CompletedProcess(command, 0, stdout="ok\n")

    monkeypatch.setattr("orcheo_sdk.cli.plugin.subprocess.run", fake_run)
    result = plugin_module._run_stack_subprocess(["docker", "ps"])
    assert result.returncode == 0


def test_run_stack_plugin_passthrough_prints_to_console(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    state = _make_cli_state(human=True)
    monkeypatch.setattr(
        plugin_module,
        "_stack_plugin_command",
        lambda *, args, human: ["dummy"],
    )

    def fake_subprocess(
        command: list[str], *, expected_exit_codes: set[int] | None = None
    ) -> subprocess.CompletedProcess[str]:
        del expected_exit_codes
        return subprocess.CompletedProcess(command, 0, stdout="hello\n")

    monkeypatch.setattr(plugin_module, "_run_stack_subprocess", fake_subprocess)
    plugin_module._run_stack_plugin_passthrough(args=["list"], state=state)
    assert "hello" in state.console.export_text()


def test_run_stack_plugin_passthrough_uses_typer_echo(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    state = _make_cli_state(human=False)
    monkeypatch.setattr(
        plugin_module,
        "_stack_plugin_command",
        lambda *, args, human: ["dummy"],
    )

    def fake_subprocess(
        command: list[str], *, expected_exit_codes: set[int] | None = None
    ) -> subprocess.CompletedProcess[str]:
        del expected_exit_codes
        return subprocess.CompletedProcess(command, 0, stdout="machine\n")

    monkeypatch.setattr(plugin_module, "_run_stack_subprocess", fake_subprocess)
    echoed: list[str] = []

    def capture_echo(message: str) -> None:
        echoed.append(message)

    monkeypatch.setattr("typer.echo", capture_echo)
    plugin_module._run_stack_plugin_passthrough(args=["list"], state=state)
    assert echoed == ["machine"]


def test_run_stack_plugin_passthrough_doctor_raises_exit(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    state = _make_cli_state(human=False)
    monkeypatch.setattr(
        plugin_module,
        "_stack_plugin_command",
        lambda *, args, human: ["dummy"],
    )

    def fake_subprocess(
        command: list[str], *, expected_exit_codes: set[int] | None = None
    ) -> subprocess.CompletedProcess[str]:
        del expected_exit_codes
        return subprocess.CompletedProcess(command, 1, stdout="")

    monkeypatch.setattr(plugin_module, "_run_stack_subprocess", fake_subprocess)
    with pytest.raises(typer.Exit) as excinfo:
        plugin_module._run_stack_plugin_passthrough(args=["doctor"], state=state)
    assert excinfo.value.exit_code == 1


def test_install_plugin_stack_runtime_restart(monkeypatch: pytest.MonkeyPatch) -> None:
    state = MagicMock()
    state.human = False
    payload = {
        "plugin": {"name": "example"},
        "impact": {
            "change_type": "install",
            "affected_component_kinds": [],
            "affected_component_ids": [],
            "activation_mode": "silent_hot_reload",
            "prompt_required": False,
            "restart_required": True,
        },
    }
    monkeypatch.setattr(plugin_module, "_state", lambda ctx: state)
    monkeypatch.setattr(plugin_module, "_use_stack_runtime", lambda runtime: True)
    monkeypatch.setattr(plugin_module, "_is_local_plugin_ref", lambda ref: False)
    monkeypatch.setattr(plugin_module, "_run_stack_plugin_json", lambda args: payload)
    restarts: list[CLIState] = []

    def capture_restart(console_state: CLIState) -> None:
        restarts.append(console_state)

    monkeypatch.setattr(
        plugin_module, "_restart_running_stack_services", capture_restart
    )
    printed: list[dict[str, Any]] = []
    monkeypatch.setattr(
        plugin_module, "print_json", lambda value: printed.append(value)
    )
    plugin_module.install_plugin(None, "example-ref", runtime="auto")
    assert restarts == [state]
    assert printed == [payload]


def test_doctor_plugins_stack_runtime_passthrough(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    state = MagicMock()
    monkeypatch.setattr(plugin_module, "_state", lambda ctx: state)
    monkeypatch.setattr(plugin_module, "_use_stack_runtime", lambda runtime: True)
    calls: list[tuple[list[str], CLIState]] = []

    def capture_passthrough(*, args: list[str], state: CLIState) -> None:
        calls.append((args, state))

    monkeypatch.setattr(
        plugin_module,
        "_run_stack_plugin_passthrough",
        capture_passthrough,
    )
    plugin_module.doctor_plugins(None, runtime="auto")
    assert calls == [(["doctor"], state)]


def test_plugin_list_empty(runner: CliRunner, machine_env: dict[str, str]) -> None:
    """Machine mode renders empty output when no plugins are installed."""
    result = runner.invoke(app, ["plugin", "list"], env=machine_env)
    assert result.exit_code == 0
    assert result.stdout.strip() == "(empty)"


def test_plugin_install_list_and_show(
    runner: CliRunner, machine_env: dict[str, str], tmp_path: Path
) -> None:
    """Install a local plugin fixture and expose it through list/show."""
    fixture_path = _copy_fixture(tmp_path, "node_plugin")

    install_result = runner.invoke(
        app, ["plugin", "install", str(fixture_path)], env=machine_env
    )
    assert install_result.exit_code == 0
    install_payload = json.loads(install_result.stdout)
    assert install_payload["plugin"]["name"] == "orcheo-plugin-fixture-node"

    list_result = runner.invoke(app, ["plugin", "list"], env=machine_env)
    assert list_result.exit_code == 0
    assert "orcheo-plugin-fixture-node" in list_result.stdout

    show_result = runner.invoke(
        app,
        ["plugin", "show", "orcheo-plugin-fixture-node"],
        env=machine_env,
    )
    assert show_result.exit_code == 0
    show_payload = json.loads(show_result.stdout)
    assert show_payload["plugin_api_version"] == 1
    assert show_payload["exports"] == ["nodes"]

    plugin_dir = _plugin_dir(machine_env)
    assert (plugin_dir / "plugins.toml").exists()
    assert (plugin_dir / "plugin-lock.toml").exists()
    assert (plugin_dir / "venv").exists()

    node_list_result = runner.invoke(app, ["node", "list"], env=machine_env)
    assert node_list_result.exit_code == 0
    assert "FixturePluginNode" in node_list_result.stdout


def test_plugin_install_machine_mode_returns_json(
    runner: CliRunner, machine_env: dict[str, str], tmp_path: Path
) -> None:
    """Machine mode returns structured install output."""
    fixture_path = _copy_fixture(tmp_path, "edge_plugin")

    result = runner.invoke(
        app, ["plugin", "install", str(fixture_path)], env=machine_env
    )
    assert result.exit_code == 0
    payload = json.loads(result.stdout)
    assert payload["plugin"]["name"] == "orcheo-plugin-fixture-edge"
    assert payload["impact"]["activation_mode"] == "silent_hot_reload"

    edge_list_result = runner.invoke(app, ["edge", "list"], env=machine_env)
    assert edge_list_result.exit_code == 0
    assert "FixturePluginEdge" in edge_list_result.stdout


def test_plugin_install_targets_stack_and_restarts_running_services(
    runner: CliRunner,
    machine_env: dict[str, str],
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
) -> None:
    """Stack-aware install shells into the backend container and restarts services."""
    stack_dir = tmp_path / "stack"
    compose_file = stack_dir / "docker-compose.yml"
    compose_file.write_text("services: {}\n", encoding="utf-8")

    executed: list[list[str]] = []
    responses = iter(
        [
            subprocess.CompletedProcess(["docker"], 0, stdout="backend\nworker\n"),
            subprocess.CompletedProcess(
                ["docker"],
                0,
                stdout=json.dumps(
                    {
                        "plugin": {"name": "orcheo-plugin-lark-listener"},
                        "impact": {
                            "change_type": "install",
                            "affected_component_kinds": ["listeners"],
                            "affected_component_ids": ["lark"],
                            "activation_mode": "restart_required",
                            "prompt_required": False,
                            "restart_required": True,
                        },
                    }
                ),
            ),
            subprocess.CompletedProcess(["docker"], 0, stdout="backend\nworker\n"),
            subprocess.CompletedProcess(["docker"], 0, stdout=""),
        ]
    )

    def _run(
        command: list[str],
        *,
        check: bool,
        capture_output: bool,
        text: bool,
    ) -> subprocess.CompletedProcess[str]:
        del check, capture_output, text
        executed.append(command)
        return next(responses)

    monkeypatch.setattr("orcheo_sdk.cli.plugin.subprocess.run", _run)
    monkeypatch.setattr(
        "orcheo_sdk.cli.plugin.shutil.which", lambda _: "/usr/bin/docker"
    )

    result = runner.invoke(
        app,
        ["plugin", "install", "orcheo-plugin-lark-listener"],
        env=_stack_env(machine_env, stack_dir),
    )

    assert result.exit_code == 0
    payload = json.loads(result.stdout)
    assert payload["plugin"]["name"] == "orcheo-plugin-lark-listener"
    assert executed[1] == [
        "docker",
        "compose",
        "-f",
        str(compose_file),
        "--project-directory",
        str(stack_dir),
        "exec",
        "-T",
        "backend",
        "orcheo",
        "plugin",
        "install",
        "orcheo-plugin-lark-listener",
        "--runtime",
        "local",
    ]
    assert executed[3] == [
        "docker",
        "compose",
        "-f",
        str(compose_file),
        "--project-directory",
        str(stack_dir),
        "restart",
        "backend",
        "worker",
    ]


def test_plugin_install_git_ref_targets_stack_runtime(
    runner: CliRunner,
    machine_env: dict[str, str],
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
) -> None:
    """Git refs should install into the stack when stack runtime is requested."""
    stack_dir = tmp_path / "stack"
    compose_file = stack_dir / "docker-compose.yml"
    compose_file.write_text("services: {}\n", encoding="utf-8")

    git_ref = "git+https://github.com/ShaojieJiang/orcheo-plugin-wecom-listener.git"
    executed: list[list[str]] = []
    responses = iter(
        [
            subprocess.CompletedProcess(["docker"], 0, stdout="backend\n"),
            subprocess.CompletedProcess(
                ["docker"],
                0,
                stdout=json.dumps(
                    {
                        "plugin": {"name": "orcheo-plugin-wecom-listener"},
                        "impact": {
                            "change_type": "install",
                            "affected_component_kinds": ["listeners"],
                            "affected_component_ids": ["wecom"],
                            "activation_mode": "restart_required",
                            "prompt_required": False,
                            "restart_required": True,
                        },
                    }
                ),
            ),
            subprocess.CompletedProcess(["docker"], 0, stdout="backend\n"),
            subprocess.CompletedProcess(["docker"], 0, stdout=""),
        ]
    )

    def _run(
        command: list[str],
        *,
        check: bool,
        capture_output: bool,
        text: bool,
    ) -> subprocess.CompletedProcess[str]:
        del check, capture_output, text
        executed.append(command)
        return next(responses)

    monkeypatch.setattr("orcheo_sdk.cli.plugin.subprocess.run", _run)
    monkeypatch.setattr(
        "orcheo_sdk.cli.plugin.shutil.which", lambda _: "/usr/bin/docker"
    )

    result = runner.invoke(
        app,
        ["plugin", "install", "--runtime", "stack", git_ref],
        env=_stack_env(machine_env, stack_dir),
    )

    assert result.exit_code == 0
    payload = json.loads(result.stdout)
    assert payload["plugin"]["name"] == "orcheo-plugin-wecom-listener"
    assert executed[1] == [
        "docker",
        "compose",
        "-f",
        str(compose_file),
        "--project-directory",
        str(stack_dir),
        "exec",
        "-T",
        "backend",
        "orcheo",
        "plugin",
        "install",
        git_ref,
        "--runtime",
        "local",
    ]


def test_plugin_install_local_path_uses_stack_runtime(
    runner: CliRunner,
    machine_env: dict[str, str],
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
) -> None:
    """Local path installs should stage into the stack runtime when selected."""
    fixture_path = _copy_fixture(tmp_path, "node_plugin")
    stack_dir = tmp_path / "stack"
    stack_dir.mkdir(exist_ok=True)
    compose_file = stack_dir / "docker-compose.yml"
    compose_file.write_text("services: {}\n", encoding="utf-8")

    monkeypatch.setattr(
        "orcheo_sdk.cli.plugin._copy_local_plugin_ref_into_stack",
        lambda ref: "/data/plugin-sources/abc/node_plugin",
    )
    stack_install = MagicMock(
        return_value={
            "plugin": {"name": "orcheo-plugin-fixture-node"},
            "impact": {
                "change_type": "install",
                "affected_component_kinds": [],
                "affected_component_ids": [],
                "activation_mode": "restart_required",
                "prompt_required": False,
                "restart_required": False,
            },
        }
    )
    monkeypatch.setattr("orcheo_sdk.cli.plugin._run_stack_plugin_json", stack_install)

    result = runner.invoke(
        app,
        ["plugin", "install", str(fixture_path), "--runtime", "stack"],
        env=_stack_env(machine_env, stack_dir),
    )

    assert result.exit_code == 0
    payload = json.loads(result.stdout)
    assert payload["plugin"]["name"] == "orcheo-plugin-fixture-node"
    stack_install.assert_called_once_with(
        ["install", "/data/plugin-sources/abc/node_plugin"]
    )


def test_plugin_update_prompts_for_existing_hot_reloadable_plugin(
    runner: CliRunner, env: dict[str, str], tmp_path: Path
) -> None:
    """Updating an existing node plugin prompts for confirmation."""
    fixture_path = _copy_fixture(tmp_path, "node_plugin")
    result = runner.invoke(app, ["plugin", "install", str(fixture_path)], env=env)
    assert result.exit_code == 0

    pyproject_path = fixture_path / "pyproject.toml"
    pyproject_path.write_text(
        pyproject_path.read_text(encoding="utf-8").replace(
            'version = "0.1.0"', 'version = "0.2.0"'
        ),
        encoding="utf-8",
    )

    update_result = runner.invoke(
        app,
        ["plugin", "update", "orcheo-plugin-fixture-node"],
        env=env,
        input="y\n",
    )
    assert update_result.exit_code == 0
    assert "Update orcheo-plugin-fixture-node" in update_result.stdout

    show_result = runner.invoke(
        app,
        ["plugin", "show", "orcheo-plugin-fixture-node"],
        env=env,
    )
    assert show_result.exit_code == 0
    assert "0.2.0" in show_result.stdout


def test_plugin_disable_and_enable(
    runner: CliRunner, env: dict[str, str], tmp_path: Path
) -> None:
    """Disable and re-enable a listener plugin."""
    fixture_path = _copy_fixture(tmp_path, "listener_plugin")
    result = runner.invoke(app, ["plugin", "install", str(fixture_path)], env=env)
    assert result.exit_code == 0

    disable_result = runner.invoke(
        app,
        ["plugin", "disable", "orcheo-plugin-fixture-listener"],
        env=env,
        input="y\n",
    )
    assert disable_result.exit_code == 0
    assert "Disable orcheo-plugin-fixture-listener" in disable_result.stdout

    show_disabled = runner.invoke(
        app,
        ["plugin", "show", "orcheo-plugin-fixture-listener"],
        env=env,
    )
    assert show_disabled.exit_code == 0
    assert "'enabled': False" in show_disabled.stdout

    enable_result = runner.invoke(
        app,
        ["plugin", "enable", "orcheo-plugin-fixture-listener", "--force"],
        env=env,
    )
    assert enable_result.exit_code == 0

    show_enabled = runner.invoke(
        app,
        ["plugin", "show", "orcheo-plugin-fixture-listener"],
        env=env,
    )
    assert show_enabled.exit_code == 0
    assert "'enabled': True" in show_enabled.stdout


def test_plugin_listener_registration_is_available_to_compiler(
    runner: CliRunner,
    machine_env: dict[str, str],
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
) -> None:
    """Installing a listener plugin makes its platform available to compilation."""
    from uuid import uuid4
    from orcheo.listeners.compiler import compile_listener_subscriptions
    from orcheo.plugins import reset_plugin_loader_for_tests

    fixture_path = _copy_fixture(tmp_path, "listener_plugin")
    result = runner.invoke(
        app, ["plugin", "install", str(fixture_path)], env=machine_env
    )
    assert result.exit_code == 0

    for key, value in machine_env.items():
        monkeypatch.setenv(key, value)

    reset_plugin_loader_for_tests()
    subscriptions = compile_listener_subscriptions(
        uuid4(),
        uuid4(),
        {
            "index": {
                "listeners": [
                    {
                        "node_name": "fixture",
                        "platform": "fixture-listener",
                        "credential_ref": "fixture-credential",
                    }
                ]
            }
        },
    )
    assert len(subscriptions) == 1
    assert subscriptions[0].platform == "fixture-listener"
    assert subscriptions[0].bot_identity_key == "fixture-listener:fixture-credential"


def test_plugin_uninstall_rebuilds_environment(
    runner: CliRunner, machine_env: dict[str, str], tmp_path: Path
) -> None:
    """Uninstall removes the plugin from desired and resolved state."""
    fixture_path = _copy_fixture(tmp_path, "node_plugin")
    result = runner.invoke(
        app, ["plugin", "install", str(fixture_path)], env=machine_env
    )
    assert result.exit_code == 0

    uninstall_result = runner.invoke(
        app,
        ["plugin", "uninstall", "orcheo-plugin-fixture-node", "--force"],
        env=machine_env,
    )
    assert uninstall_result.exit_code == 0
    payload = json.loads(uninstall_result.stdout)
    assert payload["name"] == "orcheo-plugin-fixture-node"

    list_result = runner.invoke(app, ["plugin", "list"], env=machine_env)
    assert list_result.exit_code == 0
    assert list_result.stdout.strip() == "(empty)"


def test_plugin_uninstall_prompt_happens_before_mutation(
    runner: CliRunner, env: dict[str, str], tmp_path: Path
) -> None:
    """Declining uninstall confirmation should leave plugin state unchanged."""
    fixture_path = _copy_fixture(tmp_path, "node_plugin")
    install_result = runner.invoke(
        app, ["plugin", "install", str(fixture_path)], env=env
    )
    assert install_result.exit_code == 0

    uninstall_result = runner.invoke(
        app,
        ["plugin", "uninstall", "orcheo-plugin-fixture-node"],
        env=env,
        input="n\n",
    )
    assert uninstall_result.exit_code == 1

    show_result = runner.invoke(
        app,
        ["plugin", "show", "orcheo-plugin-fixture-node"],
        env=env,
    )
    assert show_result.exit_code == 0


def test_plugin_update_prompt_happens_before_mutation(
    runner: CliRunner, env: dict[str, str], tmp_path: Path
) -> None:
    """Declining update confirmation should not rebuild to newer plugin version."""
    fixture_path = _copy_fixture(tmp_path, "node_plugin")
    install_result = runner.invoke(
        app, ["plugin", "install", str(fixture_path)], env=env
    )
    assert install_result.exit_code == 0

    pyproject_path = fixture_path / "pyproject.toml"
    pyproject_path.write_text(
        pyproject_path.read_text(encoding="utf-8").replace(
            'version = "0.1.0"', 'version = "0.2.0"'
        ),
        encoding="utf-8",
    )

    update_result = runner.invoke(
        app,
        ["plugin", "update", "orcheo-plugin-fixture-node"],
        env=env,
        input="n\n",
    )
    assert update_result.exit_code == 1

    show_result = runner.invoke(
        app,
        ["plugin", "show", "orcheo-plugin-fixture-node"],
        env=env,
    )
    assert show_result.exit_code == 0
    assert "0.1.0" in show_result.stdout


def test_plugin_doctor_reports_broken_plugin(
    runner: CliRunner, machine_env: dict[str, str], tmp_path: Path
) -> None:
    """Doctor surfaces import failures as errors."""
    fixture_path = _copy_fixture(tmp_path, "broken_plugin")
    install_result = runner.invoke(
        app,
        ["plugin", "install", str(fixture_path)],
        env=machine_env,
    )
    assert install_result.exit_code == 0

    doctor_result = runner.invoke(app, ["plugin", "doctor"], env=machine_env)
    assert doctor_result.exit_code == 1
    payload = json.loads(doctor_result.stdout)
    assert payload["has_errors"] is True
    assert any("failed to import" in check["message"] for check in payload["checks"])


def test_plugin_install_incompatible_manifest_is_transactional(
    runner: CliRunner, machine_env: dict[str, str], tmp_path: Path
) -> None:
    """Failed installs do not leave desired state or an activated plugin venv behind."""
    fixture_path = _copy_fixture(tmp_path, "incompatible_plugin")
    result = runner.invoke(
        app, ["plugin", "install", str(fixture_path)], env=machine_env
    )
    assert result.exit_code != 0

    plugin_dir = _plugin_dir(machine_env)
    state_file = plugin_dir / "plugins.toml"
    lock_file = plugin_dir / "plugin-lock.toml"

    assert not state_file.exists() or not state_file.read_text(encoding="utf-8").strip()
    assert not lock_file.exists() or not lock_file.read_text(encoding="utf-8").strip()

    list_result = runner.invoke(app, ["plugin", "list"], env=machine_env)
    assert list_result.exit_code == 0
    assert list_result.stdout.strip() == "(empty)"


def test_plugin_list_human_mode_render_table(
    runner: CliRunner, env: dict[str, str]
) -> None:
    """Human mode renders a table instead of machine output (line 91)."""
    mock_rows = [
        {
            "name": "orcheo-plugin-fixture-node",
            "enabled": True,
            "status": "installed",
            "version": "0.1.0",
            "exports": ["nodes"],
            "source": "/path/to/plugin",
        }
    ]
    with patch("orcheo_sdk.cli.plugin.list_plugins_data", return_value=mock_rows):
        result = runner.invoke(app, ["plugin", "list"], env=env)
    assert result.exit_code == 0
    # Human mode shows a rich table (machine mode shows markdown)
    assert "Installed Plugins" in result.stdout
    assert "orcheo-plugin-f" in result.stdout


def test_plugin_show_error_nonexistent(
    runner: CliRunner, machine_env: dict[str, str]
) -> None:
    """show_plugin raises BadParameter when plugin not found (lines 116-117)."""
    result = runner.invoke(
        app, ["plugin", "show", "nonexistent-plugin"], env=machine_env
    )
    assert result.exit_code != 0


def test_plugin_update_no_name_no_all_raises(
    runner: CliRunner, machine_env: dict[str, str]
) -> None:
    """update with no name and no --all raises BadParameter (line 206)."""
    result = runner.invoke(app, ["plugin", "update"], env=machine_env)
    assert result.exit_code != 0
    assert "Provide a plugin name" in result.output


def test_plugin_update_single_preview_error(
    runner: CliRunner, machine_env: dict[str, str]
) -> None:
    """update single raises BadParameter on preview PluginError (lines 209-210)."""
    result = runner.invoke(
        app, ["plugin", "update", "nonexistent-plugin"], env=machine_env
    )
    assert result.exit_code != 0


def test_plugin_update_single_machine_mode_output(
    runner: CliRunner, machine_env: dict[str, str]
) -> None:
    """update single in machine mode prints JSON and returns (lines 226-232)."""
    mock_impact = MagicMock()
    mock_impact.prompt_required = False
    mock_impact.activation_mode = "silent_hot_reload"
    mock_preview = {"name": "my-plugin", "impact": mock_impact}
    mock_plugin = {"name": "my-plugin", "version": "0.2.0"}
    mock_payload = {"plugin": mock_plugin, "impact": mock_impact}

    with patch(
        "orcheo_sdk.cli.plugin.preview_update_plugin_data",
        return_value=mock_preview,
    ):
        with patch(
            "orcheo_sdk.cli.plugin.update_plugin_data",
            return_value=mock_payload,
        ):
            result = runner.invoke(
                app,
                ["plugin", "update", "my-plugin", "--force"],
                env=machine_env,
            )
    assert result.exit_code == 0
    payload = json.loads(result.stdout)
    assert payload["plugin"]["name"] == "my-plugin"
    assert "impact" in payload


def test_plugin_update_single_error_in_update_data(
    runner: CliRunner, machine_env: dict[str, str]
) -> None:
    """update_plugin_data raises BadParameter (lines 223-224)."""
    from orcheo.plugins.manager import PluginError

    mock_impact = MagicMock()
    mock_impact.prompt_required = False
    mock_impact.activation_mode = "silent_hot_reload"
    mock_preview = {"name": "my-plugin", "impact": mock_impact}

    with patch(
        "orcheo_sdk.cli.plugin.preview_update_plugin_data",
        return_value=mock_preview,
    ):
        with patch(
            "orcheo_sdk.cli.plugin.update_plugin_data",
            side_effect=PluginError("update failed"),
        ):
            result = runner.invoke(
                app,
                ["plugin", "update", "my-plugin", "--force"],
                env=machine_env,
            )
    assert result.exit_code != 0


def test_plugin_update_all_no_plugins_machine_mode(
    runner: CliRunner, machine_env: dict[str, str]
) -> None:
    """update --all with no installed plugins outputs empty JSON (lines 181-191)."""
    with patch(
        "orcheo_sdk.cli.plugin.preview_update_all_plugins_data",
        return_value=[],
    ):
        with patch(
            "orcheo_sdk.cli.plugin.update_all_plugins_data",
            return_value=[],
        ):
            result = runner.invoke(
                app, ["plugin", "update", "--all", "--force"], env=machine_env
            )
    assert result.exit_code == 0
    payload = json.loads(result.stdout)
    assert payload == []


def test_plugin_update_all_human_mode(runner: CliRunner, env: dict[str, str]) -> None:
    """update --all in human mode uses render_json (lines 192-202)."""
    mock_impact = MagicMock()
    mock_impact.prompt_required = False
    mock_impact.activation_mode = "silent_hot_reload"
    mock_preview = [{"name": "my-plugin", "impact": mock_impact}]
    mock_plugin = {"name": "my-plugin", "version": "0.2.0"}
    mock_payload = [{"plugin": mock_plugin, "impact": mock_impact}]

    with patch(
        "orcheo_sdk.cli.plugin.preview_update_all_plugins_data",
        return_value=mock_preview,
    ):
        with patch(
            "orcheo_sdk.cli.plugin.update_all_plugins_data",
            return_value=mock_payload,
        ):
            result = runner.invoke(
                app, ["plugin", "update", "--all", "--force"], env=env
            )
    assert result.exit_code == 0


def test_plugin_update_all_preview_error(
    runner: CliRunner, machine_env: dict[str, str]
) -> None:
    """update --all raises BadParameter on preview PluginError (lines 165-166)."""
    from orcheo.plugins.manager import PluginError

    with patch(
        "orcheo_sdk.cli.plugin.preview_update_all_plugins_data",
        side_effect=PluginError("preview failed"),
    ):
        result = runner.invoke(
            app, ["plugin", "update", "--all", "--force"], env=machine_env
        )
    assert result.exit_code != 0


def test_plugin_update_all_update_error(
    runner: CliRunner, machine_env: dict[str, str]
) -> None:
    """update --all raises BadParameter on update_all_plugins_data PluginError."""
    from orcheo.plugins.manager import PluginError

    mock_impact = MagicMock()
    mock_impact.prompt_required = False
    mock_impact.activation_mode = "silent_hot_reload"
    mock_preview = [{"name": "some-plugin", "impact": mock_impact}]

    with patch(
        "orcheo_sdk.cli.plugin.preview_update_all_plugins_data",
        return_value=mock_preview,
    ):
        with patch(
            "orcheo_sdk.cli.plugin.update_all_plugins_data",
            side_effect=PluginError("update all failed"),
        ):
            result = runner.invoke(
                app, ["plugin", "update", "--all", "--force"], env=machine_env
            )
    assert result.exit_code != 0


def test_plugin_uninstall_preview_error(
    runner: CliRunner, machine_env: dict[str, str]
) -> None:
    """uninstall raises BadParameter on preview PluginError (lines 250-251)."""
    result = runner.invoke(
        app, ["plugin", "uninstall", "nonexistent-plugin", "--force"], env=machine_env
    )
    assert result.exit_code != 0


def test_plugin_uninstall_data_error(
    runner: CliRunner, machine_env: dict[str, str]
) -> None:
    """uninstall_plugin_data raises BadParameter (lines 261-262)."""
    from orcheo.plugins.manager import PluginError

    mock_impact = MagicMock()
    mock_impact.prompt_required = False
    mock_impact.activation_mode = "silent_hot_reload"
    mock_preview = {"name": "my-plugin", "impact": mock_impact}

    with patch(
        "orcheo_sdk.cli.plugin.preview_uninstall_plugin_data",
        return_value=mock_preview,
    ):
        with patch(
            "orcheo_sdk.cli.plugin.uninstall_plugin_data",
            side_effect=PluginError("uninstall failed"),
        ):
            result = runner.invoke(
                app,
                ["plugin", "uninstall", "my-plugin", "--force"],
                env=machine_env,
            )
    assert result.exit_code != 0


def test_plugin_uninstall_human_mode_output(
    runner: CliRunner, env: dict[str, str]
) -> None:
    """Uninstall in human mode prints confirmation message (lines 267-268)."""
    mock_impact = MagicMock()
    mock_impact.prompt_required = False
    mock_impact.activation_mode = "silent_hot_reload"
    mock_preview = {"name": "my-plugin", "impact": mock_impact}
    mock_payload = {"name": "my-plugin", "impact": mock_impact}

    with patch(
        "orcheo_sdk.cli.plugin.preview_uninstall_plugin_data",
        return_value=mock_preview,
    ):
        with patch(
            "orcheo_sdk.cli.plugin.uninstall_plugin_data",
            return_value=mock_payload,
        ):
            result = runner.invoke(
                app,
                ["plugin", "uninstall", "my-plugin", "--force"],
                env=env,
            )
    assert result.exit_code == 0
    assert "my-plugin" in result.output


def test_plugin_enable_preview_error(
    runner: CliRunner, machine_env: dict[str, str]
) -> None:
    """enable raises BadParameter on preview PluginError (lines 284-285)."""
    result = runner.invoke(
        app, ["plugin", "enable", "nonexistent-plugin", "--force"], env=machine_env
    )
    assert result.exit_code != 0


def test_plugin_enable_data_error(
    runner: CliRunner, machine_env: dict[str, str]
) -> None:
    """enable_plugin_data raises BadParameter (lines 295-296)."""
    from orcheo.plugins.manager import PluginError

    mock_impact = MagicMock()
    mock_impact.prompt_required = False
    mock_impact.activation_mode = "silent_hot_reload"
    mock_preview = {"name": "my-plugin", "impact": mock_impact}

    with patch(
        "orcheo_sdk.cli.plugin.preview_enable_plugin_data",
        return_value=mock_preview,
    ):
        with patch(
            "orcheo_sdk.cli.plugin.enable_plugin_data",
            side_effect=PluginError("enable failed"),
        ):
            result = runner.invoke(
                app,
                ["plugin", "enable", "my-plugin", "--force"],
                env=machine_env,
            )
    assert result.exit_code != 0


def test_plugin_enable_machine_mode_output(
    runner: CliRunner, machine_env: dict[str, str]
) -> None:
    """Enable in machine mode prints JSON (lines 299-300)."""
    mock_impact = MagicMock()
    mock_impact.prompt_required = False
    mock_impact.activation_mode = "silent_hot_reload"
    mock_preview = {"name": "my-plugin", "impact": mock_impact}
    mock_payload = {"name": "my-plugin", "impact": mock_impact}

    with patch(
        "orcheo_sdk.cli.plugin.preview_enable_plugin_data",
        return_value=mock_preview,
    ):
        with patch(
            "orcheo_sdk.cli.plugin.enable_plugin_data",
            return_value=mock_payload,
        ):
            result = runner.invoke(
                app,
                ["plugin", "enable", "my-plugin", "--force"],
                env=machine_env,
            )
    assert result.exit_code == 0
    payload = json.loads(result.stdout)
    assert payload["name"] == "my-plugin"


def test_plugin_disable_preview_error(
    runner: CliRunner, machine_env: dict[str, str]
) -> None:
    """disable raises BadParameter on preview PluginError (lines 318-319)."""
    result = runner.invoke(
        app, ["plugin", "disable", "nonexistent-plugin", "--force"], env=machine_env
    )
    assert result.exit_code != 0


def test_plugin_disable_data_error(
    runner: CliRunner, machine_env: dict[str, str]
) -> None:
    """disable_plugin_data raises BadParameter (lines 329-330)."""
    from orcheo.plugins.manager import PluginError

    mock_impact = MagicMock()
    mock_impact.prompt_required = False
    mock_impact.activation_mode = "silent_hot_reload"
    mock_preview = {"name": "my-plugin", "impact": mock_impact}

    with patch(
        "orcheo_sdk.cli.plugin.preview_disable_plugin_data",
        return_value=mock_preview,
    ):
        with patch(
            "orcheo_sdk.cli.plugin.disable_plugin_data",
            side_effect=PluginError("disable failed"),
        ):
            result = runner.invoke(
                app,
                ["plugin", "disable", "my-plugin", "--force"],
                env=machine_env,
            )
    assert result.exit_code != 0


def test_plugin_disable_machine_mode_output(
    runner: CliRunner, machine_env: dict[str, str]
) -> None:
    """Disable in machine mode prints JSON (lines 333-334)."""
    mock_impact = MagicMock()
    mock_impact.prompt_required = False
    mock_impact.activation_mode = "silent_hot_reload"
    mock_preview = {"name": "my-plugin", "impact": mock_impact}
    mock_payload = {"name": "my-plugin", "impact": mock_impact}

    with patch(
        "orcheo_sdk.cli.plugin.preview_disable_plugin_data",
        return_value=mock_preview,
    ):
        with patch(
            "orcheo_sdk.cli.plugin.disable_plugin_data",
            return_value=mock_payload,
        ):
            result = runner.invoke(
                app,
                ["plugin", "disable", "my-plugin", "--force"],
                env=machine_env,
            )
    assert result.exit_code == 0
    payload = json.loads(result.stdout)
    assert payload["name"] == "my-plugin"


def test_plugin_doctor_human_mode_with_errors(
    runner: CliRunner, env: dict[str, str]
) -> None:
    """Doctor in human mode renders a table and exits 1 on errors (lines 348-360)."""
    mock_payload = {
        "has_errors": True,
        "checks": [
            {
                "name": "some-check",
                "severity": "ERROR",
                "ok": False,
                "message": "Something failed",
            }
        ],
    }
    with patch("orcheo_sdk.cli.plugin.doctor_plugins_data", return_value=mock_payload):
        result = runner.invoke(app, ["plugin", "doctor"], env=env)
    assert result.exit_code == 1
    assert "Plugin Doctor" in result.output


def test_plugin_doctor_human_mode_no_errors(
    runner: CliRunner, env: dict[str, str]
) -> None:
    """Doctor in human mode exits 0 when no errors (line 358 False branch)."""
    mock_payload = {
        "has_errors": False,
        "checks": [
            {
                "name": "some-check",
                "severity": "INFO",
                "ok": True,
                "message": "All good",
            }
        ],
    }
    with patch("orcheo_sdk.cli.plugin.doctor_plugins_data", return_value=mock_payload):
        result = runner.invoke(app, ["plugin", "doctor"], env=env)
    assert result.exit_code == 0
    assert "Plugin Doctor" in result.output


def test_plugin_doctor_machine_mode_no_errors(
    runner: CliRunner, machine_env: dict[str, str]
) -> None:
    """Doctor in machine mode exits 0 and returns when no errors (line 348)."""
    mock_payload = {
        "has_errors": False,
        "checks": [
            {
                "name": "some-check",
                "severity": "INFO",
                "ok": True,
                "message": "All good",
            }
        ],
    }
    with patch("orcheo_sdk.cli.plugin.doctor_plugins_data", return_value=mock_payload):
        result = runner.invoke(app, ["plugin", "doctor"], env=machine_env)
    assert result.exit_code == 0
    payload = json.loads(result.stdout)
    assert payload["has_errors"] is False


def test_maybe_confirm_machine_mode_prompt_required_raises(
    runner: CliRunner, machine_env: dict[str, str]
) -> None:
    """_maybe_confirm raises BadParameter in machine mode when prompt is needed."""
    mock_impact = MagicMock()
    mock_impact.prompt_required = True
    mock_impact.activation_mode = "requires_restart"
    mock_preview = {"name": "my-plugin", "impact": mock_impact}

    with patch(
        "orcheo_sdk.cli.plugin.preview_update_plugin_data",
        return_value=mock_preview,
    ):
        # No --force in machine mode with prompt_required=True → BadParameter
        result = runner.invoke(
            app,
            ["plugin", "update", "my-plugin"],
            env=machine_env,
        )
    assert result.exit_code != 0


# ---------------------------------------------------------------------------
# plugin init
# ---------------------------------------------------------------------------


def test_plugin_init_machine_mode(
    runner: CliRunner, machine_env: dict[str, str], tmp_path: Path
) -> None:
    """init in machine mode outputs JSON with path and name."""
    result = runner.invoke(
        app,
        ["plugin", "init", "hello", "--output-dir", str(tmp_path)],
        env=machine_env,
    )
    assert result.exit_code == 0
    payload = json.loads(result.stdout)
    assert payload["name"] == "orcheo-plugin-hello"
    assert Path(payload["path"]).is_dir()


def test_plugin_init_human_mode(
    runner: CliRunner, env: dict[str, str], tmp_path: Path
) -> None:
    """init in human mode prints a success message."""
    result = runner.invoke(
        app,
        ["plugin", "init", "hello", "--output-dir", str(tmp_path)],
        env=env,
    )
    assert result.exit_code == 0
    assert "orcheo-plugin-hello" in result.output


def test_plugin_init_creates_files(
    runner: CliRunner, machine_env: dict[str, str], tmp_path: Path
) -> None:
    """init creates a complete plugin skeleton."""
    result = runner.invoke(
        app,
        ["plugin", "init", "acme", "--output-dir", str(tmp_path)],
        env=machine_env,
    )
    assert result.exit_code == 0
    root = tmp_path / "orcheo-plugin-acme"
    assert (root / "pyproject.toml").is_file()
    assert (root / "src" / "orcheo_plugin_acme" / "__init__.py").is_file()
    assert (root / "tests" / "test_plugin.py").is_file()


def test_plugin_init_with_author_and_description(
    runner: CliRunner, machine_env: dict[str, str], tmp_path: Path
) -> None:
    """init passes author and description through to the scaffold."""
    result = runner.invoke(
        app,
        [
            "plugin",
            "init",
            "acme",
            "--author",
            "Jane Doe",
            "--description",
            "My cool plugin",
            "--output-dir",
            str(tmp_path),
        ],
        env=machine_env,
    )
    assert result.exit_code == 0
    pyproject = (tmp_path / "orcheo-plugin-acme" / "pyproject.toml").read_text(
        encoding="utf-8"
    )
    assert "Jane Doe" in pyproject
    assert "My cool plugin" in pyproject


def test_plugin_init_with_custom_exports(
    runner: CliRunner, machine_env: dict[str, str], tmp_path: Path
) -> None:
    """init passes comma-separated exports to the manifest."""
    result = runner.invoke(
        app,
        [
            "plugin",
            "init",
            "acme",
            "--exports",
            "nodes,edges",
            "--output-dir",
            str(tmp_path),
        ],
        env=machine_env,
    )
    assert result.exit_code == 0
    manifest = (
        tmp_path
        / "orcheo-plugin-acme"
        / "src"
        / "orcheo_plugin_acme"
        / "orcheo_plugin.toml"
    ).read_text(encoding="utf-8")
    assert '"nodes"' in manifest
    assert '"edges"' in manifest


def test_plugin_init_invalid_name(
    runner: CliRunner, machine_env: dict[str, str], tmp_path: Path
) -> None:
    """init rejects invalid plugin names."""
    result = runner.invoke(
        app,
        ["plugin", "init", "Bad-Name", "--output-dir", str(tmp_path)],
        env=machine_env,
    )
    assert result.exit_code != 0


def test_plugin_init_duplicate_directory(
    runner: CliRunner, machine_env: dict[str, str], tmp_path: Path
) -> None:
    """init rejects when target directory already exists."""
    result1 = runner.invoke(
        app,
        ["plugin", "init", "hello", "--output-dir", str(tmp_path)],
        env=machine_env,
    )
    assert result1.exit_code == 0

    result2 = runner.invoke(
        app,
        ["plugin", "init", "hello", "--output-dir", str(tmp_path)],
        env=machine_env,
    )
    assert result2.exit_code != 0



"""Sliding window and auth rate limiter unit tests."""

from __future__ import annotations
import pytest
from orcheo_backend.app.authentication import (
    AuthenticationError,
    AuthRateLimiter,
    SlidingWindowRateLimiter,
)


def test_sliding_window_rate_limiter_disabled_when_limit_zero() -> None:
    """Rate limiter does not enforce when limit is 0."""

    limiter = SlidingWindowRateLimiter(
        0, 60, code="test", message_template="Test {key}"
    )

    # Should not raise
    for _ in range(100):
        limiter.hit("test-key")


def test_sliding_window_rate_limiter_ignores_empty_key() -> None:
    """Rate limiter does not enforce when key is empty."""

    limiter = SlidingWindowRateLimiter(
        5, 60, code="test", message_template="Test {key}"
    )

    # Should not raise
    for _ in range(100):
        limiter.hit("")


def test_sliding_window_rate_limiter_reset() -> None:
    """reset() clears internal state."""

    limiter = SlidingWindowRateLimiter(
        2, 60, code="test", message_template="Test {key}"
    )

    limiter.hit("test-key")
    limiter.hit("test-key")

    limiter.reset()

    # Should not raise after reset
    limiter.hit("test-key")
    limiter.hit("test-key")


def test_auth_rate_limiter_check_ip_and_identity() -> None:
    """AuthRateLimiter checks both IP and identity limits."""

    limiter = AuthRateLimiter(ip_limit=2, identity_limit=2, interval_seconds=60)

    # IP limiting
    limiter.check_ip("1.2.3.4")
    limiter.check_ip("1.2.3.4")

    with pytest.raises(AuthenticationError) as exc:
        limiter.check_ip("1.2.3.4")
    assert exc.value.code == "auth.rate_limited.ip"

    # Identity limiting
    limiter.reset()
    limiter.check_identity("user-1")
    limiter.check_identity("user-1")

    with pytest.raises(AuthenticationError) as exc:
        limiter.check_identity("user-1")
    assert exc.value.code == "auth.rate_limited.identity"



"""Trigger configuration and dispatch helpers."""

from __future__ import annotations
from collections.abc import Mapping
from datetime import UTC, datetime
from typing import Any
from uuid import UUID
from orcheo.models.workflow import WorkflowRun
from orcheo.triggers.cron import CronTriggerConfig
from orcheo.triggers.manual import ManualDispatchRequest
from orcheo.triggers.webhook import WebhookRequest, WebhookTriggerConfig
from orcheo.vault.oauth import CredentialHealthError
from orcheo_backend.app.repository import (
    CronTriggerNotFoundError,
    WorkflowVersionNotFoundError,
)
from orcheo_backend.app.repository_postgres._base import logger
from orcheo_backend.app.repository_postgres._persistence import PostgresPersistenceMixin


def _enqueue_run_for_execution(run: WorkflowRun) -> None:
    """Enqueue a Celery task to execute the workflow run.

    This function is best-effort: if Celery/Redis is unavailable,
    the run remains pending and can be retried manually.

    NOTE: This must only be called AFTER the run has been committed to the database.
    """
    try:
        from orcheo_backend.worker.tasks import execute_run

        execute_run.delay(str(run.id))
        logger.info("Enqueued run %s for execution", run.id)
    except Exception as exc:
        logger.warning(
            "Failed to enqueue run %s for execution: %s. "
            "Run will remain pending until manually retried.",
            run.id,
            exc,
        )


class TriggerRepositoryMixin(PostgresPersistenceMixin):
    """Coordinate trigger configuration and dispatch flows."""

    async def configure_webhook_trigger(
        self,
        workflow_id: UUID,
        config: WebhookTriggerConfig,
    ) -> WebhookTriggerConfig:
        await self._ensure_initialized()
        async with self._lock:
            await self._get_workflow_locked(workflow_id)
            normalized = self._trigger_layer.configure_webhook(workflow_id, config)
            async with self._connection() as conn:
                await conn.execute(
                    """
                    INSERT INTO webhook_triggers (workflow_id, config)
                    VALUES (%s, %s)
                    ON CONFLICT(workflow_id) DO UPDATE SET config=EXCLUDED.config
                    """,
                    (str(workflow_id), self._dump_config(normalized)),
                )
            return normalized.model_copy(deep=True)

    async def get_webhook_trigger_config(
        self, workflow_id: UUID
    ) -> WebhookTriggerConfig:
        await self._ensure_initialized()
        async with self._lock:
            await self._get_workflow_locked(workflow_id)
            return self._trigger_layer.get_webhook_config(workflow_id)

    async def handle_webhook_trigger(
        self,
        workflow_id: UUID,
        *,
        method: str,
        headers: Mapping[str, str],
        query_params: Mapping[str, str],
        payload: Any,
        source_ip: str | None,
    ) -> WorkflowRun:
        await self._ensure_initialized()
        async with self._lock:
            await self._get_workflow_locked(workflow_id)
            version = await self._get_latest_version_locked(workflow_id)
            await self._ensure_workflow_health(workflow_id, actor="webhook")
            request = WebhookRequest(
                method=method,
                headers=headers,
                query_params=query_params,
                payload=payload,
                source_ip=source_ip,
            )
            dispatch = self._trigger_layer.prepare_webhook_dispatch(
                workflow_id, request
            )
            run = await self._create_run_locked(
                workflow_id=workflow_id,
                workflow_version_id=version.id,
                triggered_by=dispatch.triggered_by,
                input_payload=dispatch.input_payload,
                actor=dispatch.actor,
            )
            run_copy = run.model_copy(deep=True)
        # Enqueue AFTER lock is released to ensure commit is fully visible
        _enqueue_run_for_execution(run_copy)
        return run_copy

    async def configure_cron_trigger(
        self,
        workflow_id: UUID,
        config: CronTriggerConfig,
    ) -> CronTriggerConfig:
        await self._ensure_initialized()
        async with self._lock:
            await self._get_workflow_locked(workflow_id)
            normalized = self._trigger_layer.configure_cron(workflow_id, config)
            async with self._connection() as conn:
                await conn.execute(
                    """
                    INSERT INTO cron_triggers (workflow_id, config)
                    VALUES (%s, %s)
                    ON CONFLICT(workflow_id) DO UPDATE SET config=EXCLUDED.config
                    """,
                    (str(workflow_id), self._dump_config(normalized)),
                )
            return normalized.model_copy(deep=True)

    async def get_cron_trigger_config(self, workflow_id: UUID) -> CronTriggerConfig:
        await self._ensure_initialized()
        async with self._lock:
            await self._get_workflow_locked(workflow_id)
            config = self._trigger_layer.get_cron_config(workflow_id)
            if config is None:
                raise CronTriggerNotFoundError(  # pragma: no cover - defensive
                    f"No cron trigger configured for workflow {workflow_id}"
                )
            return config

    async def delete_cron_trigger(self, workflow_id: UUID) -> None:
        await self._ensure_initialized()
        async with self._lock:
            await self._get_workflow_locked(workflow_id)
            async with self._connection() as conn:
                await conn.execute(
                    """
                    DELETE FROM cron_triggers
                     WHERE workflow_id = %s
                    """,
                    (str(workflow_id),),
                )
            self._trigger_layer.remove_cron_config(workflow_id)

    async def dispatch_due_cron_runs(
        self, *, now: datetime | None = None
    ) -> list[WorkflowRun]:
        await self._ensure_initialized()
        reference = now or datetime.now(tz=UTC)
        if reference.tzinfo is None:
            reference = reference.replace(tzinfo=UTC)

        runs: list[WorkflowRun] = []

        async with self._lock:
            # Sync cron triggers each dispatch to reflect updates from other processes.
            await self._refresh_cron_triggers()
            plans = self._trigger_layer.collect_due_cron_dispatches(now=reference)
            for plan in plans:
                try:
                    version = await self._get_latest_version_locked(plan.workflow_id)
                except WorkflowVersionNotFoundError:
                    continue

                try:
                    await self._ensure_workflow_health(plan.workflow_id, actor="cron")
                except CredentialHealthError as exc:
                    logger.warning(
                        "Skipping cron dispatch for workflow %s due to credential "
                        "health error: %s",
                        plan.workflow_id,
                        exc,
                    )
                    continue

                run = await self._create_run_locked(
                    workflow_id=plan.workflow_id,
                    workflow_version_id=version.id,
                    triggered_by="cron",
                    input_payload={
                        "scheduled_for": plan.scheduled_for.isoformat(),
                        "timezone": plan.timezone,
                    },
                    actor="cron",
                )
                self._trigger_layer.commit_cron_dispatch(plan.workflow_id)
                # Persist the last_dispatched_at to survive worker restarts
                last_dispatched = self._trigger_layer.get_cron_last_dispatched_at(
                    plan.workflow_id
                )
                if last_dispatched is not None:  # pragma: no branch
                    async with self._connection() as conn:
                        await conn.execute(
                            """
                            UPDATE cron_triggers
                               SET last_dispatched_at = %s
                             WHERE workflow_id = %s
                            """,
                            (last_dispatched, str(plan.workflow_id)),
                        )
                runs.append(run.model_copy(deep=True))
        # Enqueue AFTER lock is released to ensure commits are fully visible
        for run in runs:
            _enqueue_run_for_execution(run)
        return runs

    async def dispatch_manual_runs(
        self, request: ManualDispatchRequest
    ) -> list[WorkflowRun]:
        await self._ensure_initialized()
        async with self._lock:
            await self._get_workflow_locked(request.workflow_id)
            try:
                latest_version = await self._get_latest_version_locked(
                    request.workflow_id
                )
            except WorkflowVersionNotFoundError as exc:
                raise WorkflowVersionNotFoundError(str(request.workflow_id)) from exc
            default_version_id = latest_version.id
            plan = self._trigger_layer.prepare_manual_dispatch(
                request, default_workflow_version_id=default_version_id
            )

            await self._ensure_workflow_health(
                request.workflow_id, actor=plan.actor or plan.triggered_by
            )

            runs: list[WorkflowRun] = []
            for resolved in plan.runs:
                version = await self._get_version_locked(resolved.workflow_version_id)
                if version.workflow_id != request.workflow_id:
                    raise WorkflowVersionNotFoundError(
                        str(resolved.workflow_version_id)
                    )

            for resolved in plan.runs:
                run = await self._create_run_locked(
                    workflow_id=request.workflow_id,
                    workflow_version_id=resolved.workflow_version_id,
                    triggered_by=plan.triggered_by,
                    input_payload=resolved.input_payload,
                    actor=plan.actor,
                )
                runs.append(run.model_copy(deep=True))
        # Enqueue AFTER lock is released to ensure commits are fully visible
        for run in runs:
            _enqueue_run_for_execution(run)
        return runs


__all__ = ["TriggerRepositoryMixin"]



"""Workflow runnable-config persistence helpers."""

from __future__ import annotations
from typing import Any
from orcheo_sdk.cli.errors import CLIError
from orcheo_sdk.cli.http import ApiClient


def _resolve_version_number(
    client: ApiClient,
    workflow_id: str,
    *,
    version: int | None,
) -> int:
    """Resolve a concrete version number for config-only updates."""
    if version is not None:
        return version
    versions = client.get(f"/api/workflows/{workflow_id}/versions")
    if not versions:
        msg = (
            f"Workflow '{workflow_id}' has no versions. "
            "Ingest a Python LangGraph script first."
        )
        raise CLIError(msg)
    latest = max(versions, key=lambda entry: entry.get("version", 0))
    resolved = latest.get("version")
    if not isinstance(resolved, int):
        msg = f"Workflow '{workflow_id}' returned an invalid version payload."
        raise CLIError(msg)
    return resolved


def save_workflow_runnable_config_data(
    client: ApiClient,
    workflow_id: str,
    *,
    runnable_config: dict[str, Any] | None,
    actor: str,
    version: int | None = None,
) -> dict[str, Any]:
    """Persist runnable config on an existing workflow version."""
    version_number = _resolve_version_number(client, workflow_id, version=version)
    payload = {
        "runnable_config": runnable_config,
        "actor": actor,
    }
    try:
        updated_version = client.put(
            f"/api/workflows/{workflow_id}/versions/{version_number}/runnable-config",
            json_body=payload,
        )
    except Exception as exc:
        raise CLIError(
            f"Failed to save runnable config for workflow '{workflow_id}' "
            f"version {version_number}: {exc}"
        ) from exc
    return {
        "workflow_id": workflow_id,
        "version": version_number,
        "runnable_config": updated_version.get("runnable_config"),
        "updated_version": updated_version,
    }


__all__ = ["save_workflow_runnable_config_data"]



"""Workflow run persistence and lifecycle helpers."""

from __future__ import annotations
from collections.abc import Callable
from typing import Any
from uuid import UUID
from orcheo.models.workflow import WorkflowRun
from orcheo_backend.app.repository import WorkflowNotFoundError
from orcheo_backend.app.repository_postgres._persistence import PostgresPersistenceMixin


class WorkflowRunMixin(PostgresPersistenceMixin):
    """Create and update workflow runs."""

    async def create_run(
        self,
        workflow_id: UUID,
        *,
        workflow_version_id: UUID,
        triggered_by: str,
        input_payload: dict[str, Any],
        actor: str | None = None,
        runnable_config: dict[str, Any] | None = None,
    ) -> WorkflowRun:
        await self._ensure_initialized()
        async with self._lock:
            workflow = await self._get_workflow_locked(workflow_id)
            if workflow.is_archived:
                raise WorkflowNotFoundError(str(workflow_id))
            await self._ensure_workflow_health(workflow_id, actor=actor or triggered_by)
            run = await self._create_run_locked(
                workflow_id=workflow_id,
                workflow_version_id=workflow_version_id,
                triggered_by=triggered_by,
                input_payload=input_payload,
                actor=actor,
                runnable_config=runnable_config,
            )
            return run.model_copy(deep=True)

    async def list_runs_for_workflow(self, workflow_id: UUID) -> list[WorkflowRun]:
        await self._ensure_initialized()
        async with self._lock:
            await self._get_workflow_locked(workflow_id)
            async with self._connection() as conn:
                cursor = await conn.execute(
                    """
                    SELECT payload
                      FROM workflow_runs
                     WHERE workflow_id = %s
                  ORDER BY created_at ASC
                    """,
                    (str(workflow_id),),
                )
                rows = await cursor.fetchall()
            result = []
            for row in rows:
                payload = row["payload"]
                if isinstance(payload, str):
                    run = WorkflowRun.model_validate_json(payload)
                else:
                    run = WorkflowRun.model_validate(payload)
                result.append(run.model_copy(deep=True))
            return result

    async def get_run(self, run_id: UUID) -> WorkflowRun:
        await self._ensure_initialized()
        async with self._lock:
            return await self._get_run_locked(run_id)

    async def mark_run_started(self, run_id: UUID, *, actor: str) -> WorkflowRun:
        return await self._update_run(run_id, lambda run: run.mark_started(actor=actor))

    async def mark_run_succeeded(
        self,
        run_id: UUID,
        *,
        actor: str,
        output: dict[str, Any] | None,
    ) -> WorkflowRun:
        run = await self._update_run(
            run_id,
            lambda candidate: candidate.mark_succeeded(actor=actor, output=output),
        )
        self._release_cron_run(run_id)
        self._trigger_layer.clear_retry_state(run_id)
        return run

    async def mark_run_failed(
        self,
        run_id: UUID,
        *,
        actor: str,
        error: str,
    ) -> WorkflowRun:
        run = await self._update_run(
            run_id,
            lambda candidate: candidate.mark_failed(actor=actor, error=error),
        )
        self._release_cron_run(run_id)
        return run

    async def mark_run_cancelled(
        self,
        run_id: UUID,
        *,
        actor: str,
        reason: str | None,
    ) -> WorkflowRun:
        run = await self._update_run(
            run_id,
            lambda candidate: candidate.mark_cancelled(actor=actor, reason=reason),
        )
        self._release_cron_run(run_id)
        self._trigger_layer.clear_retry_state(run_id)
        return run

    async def reset(self) -> None:
        await self._ensure_initialized()
        async with self._lock:
            async with self._connection() as conn:
                await conn.execute("DELETE FROM workflow_runs")
                await conn.execute("DELETE FROM workflow_versions")
                await conn.execute("DELETE FROM workflows")
                await conn.execute("DELETE FROM webhook_triggers")
                await conn.execute("DELETE FROM cron_triggers")
                await conn.execute("DELETE FROM retry_policies")
                await conn.execute("DELETE FROM listener_dedupe")
                await conn.execute("DELETE FROM listener_cursors")
                await conn.execute("DELETE FROM listener_subscriptions")
            self._trigger_layer.reset()

    async def _update_run(
        self,
        run_id: UUID,
        updater: Callable[[WorkflowRun], None],
    ) -> WorkflowRun:
        await self._ensure_initialized()
        async with self._lock:
            run = await self._get_run_locked(run_id)
            updater(run)
            async with self._connection() as conn:
                await conn.execute(
                    """
                    UPDATE workflow_runs
                       SET status = %s, payload = %s, updated_at = %s
                     WHERE id = %s
                    """,
                    (
                        run.status.value,
                        self._dump_model(run),
                        run.updated_at,
                        str(run.id),
                    ),
                )
            return run.model_copy(deep=True)


__all__ = ["WorkflowRunMixin"]



"""Integration tests between nodes and the credential resolver."""

from __future__ import annotations
import pytest
from orcheo.graph.state import State
from orcheo.nodes.logic import SetVariableNode
from orcheo.runtime.credentials import (
    CredentialResolver,
    CredentialResolverUnavailableError,
    credential_ref,
    credential_resolution,
)
from tests.runtime.credential_test_helpers import create_vault_with_secret


def test_node_decode_variables_injects_secret() -> None:
    vault = create_vault_with_secret(secret="token")
    resolver = CredentialResolver(vault)
    node = SetVariableNode(
        name="store_secret",
        variables={"token": "[[telegram_token]]"},
    )
    state = State({"results": {}, "messages": [], "inputs": {}})
    with credential_resolution(resolver):
        node.decode_variables(state)
    assert node.variables["token"] == "token"


def test_node_decode_variables_without_resolver_errors() -> None:
    node = SetVariableNode(
        name="store_secret",
        variables={"token": "[[telegram_token]]"},
    )
    state = State({"results": {}, "messages": [], "inputs": {}})
    with pytest.raises(CredentialResolverUnavailableError):
        node.decode_variables(state)


def test_node_accepts_explicit_credential_reference() -> None:
    vault = create_vault_with_secret(secret="token")
    resolver = CredentialResolver(vault)
    node = SetVariableNode(
        name="store_secret",
        variables={"token": credential_ref("telegram_token")},
    )
    state = State({"results": {}, "messages": [], "inputs": {}})
    with credential_resolution(resolver):
        node.decode_variables(state)
    assert node.variables["token"] == "token"
